#
# Parse::XMLLite - A simple, fast, XML-like parser
#
# Copyright (c) 2002, 2003 Brian Katzung, Kappa Computer Solutions, LLC.
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2002-09-09	Created by Brian Katzung
#

package Parse::XMLLite;

use strict;
use Exporter;
use vars qw(@EXPORT_OK @ISA);

sub dispatch;

@ISA = qw(Exporter);
@EXPORT_OK = qw(dispatch find_node format_attribute unparse_node);

my $a_name = '[A-Za-z_][A-Za-z0-9_]*';
my $a_value = "='[^']*'|=\"[^\"]*\"";
my $cmt = '<!--.*?-->';
my $cdta = '<!\[CDATA\[.*?\]\]>';

#----------------------------------------------------------------------
# Dispatch on a type (and possibly tag) (public function)

sub dispatch
{
	# my ($node, @table) = @_;
	my ($node, %hash) = @_;
	my $dispatch = $hash{$node->{'type'}};

	$dispatch = $dispatch->{$node->{'tag'} || ''}
	  if ($dispatch && ref($dispatch) eq 'HASH');

	if ($dispatch && ref($dispatch) eq 'CODE')
	{
		&$dispatch;
		return 1;
	}

	return 0;
}

#----------------------------------------------------------------------
# Find a node (and its parent) matching the criteria supplied as a closure
# (public function)

sub find_node
{
	my ($matcher, $node) = @_;

	foreach (@{$node->{'content'} || []})
	{
		#
		# Return content node and this node on a match.
		#
		return (wantarray? ($_, $node): $_) if ($matcher->($_));

		#
		# Recursively search content on b-tags.
		#
		if ($_->{'type'} eq 'b-tag')
		{
			my @result = find_node($matcher, $_);

			return (wantarray? @result: $result[0]) if (@result);
		}
	}

	return (wantarray? (): undef);
}

#----------------------------------------------------------------------
# Format an attribute (public function)

sub format_attribute
{
	my ($key, $value) = @_;

	$value =~ s/</\&lt;/g;
	return "$key='$value'" unless $value =~ /'/;
	return "$key=\"$value\"" unless $value =~ /"/;

	$value =~ s/"/\&quot;/g;
	return "$key=\"$value\"";
}

#----------------------------------------------------------------------
# Create a new parser (public class method)

sub new
{
	my ($class) = @_;
	my $self = bless
	{
		'tag_list' => '[A-Za-z0-9_]+',
	}, $class;

	return $self;
}

#----------------------------------------------------------------------
# Parse a string into a content tree (public instance method)

sub parse
{
	my ($self, $string) = @_;
	my $root =
	{
		'type' => 'root',
		'content' => [],
	};
	my $tags = $self->{'tag_list'};
	my @tokens = grep($_ ne '', split(
	  /($cmt|$cdta|<\/?(?i:$tags)(?:\s+$a_name$a_value)*\s*\/?>)/s,
	  $string));

	$self->_sub_parse($root, \@tokens) while (@tokens);

	return $root;
}

#----------------------------------------------------------------------
# Set parser tags (public instance method)

sub set_tags
{
	my ($self, @tags) = @_;

	$self->{'tag_list'} = join('|', @tags);
}

#----------------------------------------------------------------------
# Parse the next token or sub-container in a container
# (private instance method)
# Returns 0 at the end of the container, or 1 otherwise.

sub _sub_parse
{
	my ($self, $container, $tokens) = @_;
	my $next = shift(@$tokens);
	my $tag;

	if ($next =~ /^<!--(.*)-->$/s)			# Comment
	{
		push(@{$container->{'content'}},
		  {
		  	'type' => 'comment',
			'text' => $1,
			'parent' => $container,
		  });
		return 1;
	}

	if ($next =~ /^<!\[CDATA\[(.*)\]\]>$/s)		# CDATA
	{
		push(@{$container->{'content'}},
		  {
			'type' => 'cdata', 
			'text' => $1,
			'parent' => $container,
		  });
		return 1;
	}

	my $tags = $self->{'tag_list'};
	if ($next =~ /^<\/($tags)>$/)			# End tag
	{
		$tag = lc($1);

		if ($container->{'type'} eq 'root' ||
		  ($container->{'type'} eq 'b-tag' &&
		  $container->{'tag'} eq $tag))
		{
			# Consume the end tag if it matches the container
			# tag or the container is the root.
			push(@{$container->{'content'}},
			  {
				'type' => 'e-tag',
				'tag' => $tag,
				'parent' => $container,
			  });
			return 0;
		}

		# The end tag isn't for the current container.
		# Leave it on the token list and hope it's for a
		# parent container.
		unshift(@$tokens, $next);
		return 0;
	}

	my ($attrs, $eflag);
	($tag, $attrs, $eflag) =
	  ($next =~ /^<((?i:$tags))(\s.*[^\/])?(\/)?>/s);
	if (defined($tag))				# Tag or begin tag
	{
		my $result =
		{
			'type' => ($eflag? 'tag': 'b-tag'),
			'tag' => lc($tag),
			'parent' => $container,
		};

		# Save any attributes
		foreach (grep($_ ne '',
		  split(/\s+($a_name$a_value)/s, $attrs || '')))
		{

			next
			  unless (/^($a_name)='(.*)'$/ || /^($a_name)="(.*)"$/);

			my ($name, $value) = (lc($1), $2);

			$value =~ s/&quot;/"/g;
			$value =~ s/&lt;/</g;
			$result->{'attribute'}{$name} = $value;
		}

		push(@{$container->{'content'}}, $result);

		# Recurse for content in a begin tag
		unless ($eflag)
		{
			$result->{'content'} = [];
			while (@$tokens && $self->_sub_parse($result, $tokens))
			{
				;
			}

			my $prev;
			foreach (@{$result->{'content'}})
			{
				if ($prev)
				{
					$prev->{'next'} = $_;
					$_->{'prev'} = $prev;
				}
				$prev = $_;
			}
		}

		return 1;
	}

	push(@{$container->{'content'}},
	  {
	  	'type' => ($next =~ /^\s+$/? 'space': 'text'),
		'text' => $next,
		'parent' => $container,
	  });
	return 1;
}

#----------------------------------------------------------------------
# Recreate the tag text for a node (but not its content) (public function)

sub unparse_node
{
	my ($node, $exclude_attrs) = @_;
	my $type = $node->{'type'};

	return ' ' if ($type eq 'space');
	return $node->{'text'} if ($type eq 'text');
	return '<![CDATA[' . $node->{'text'} . ']]>' if ($type eq 'cdata');

	my $attr = $node->{'attribute'};
	my $open = ($type eq 'e-tag')? '</': '<';
	my $close = ($type eq 'tag')? '/>': '>';
	my @keys = $exclude_attrs?
	  grep($_ !~ /^(?:$exclude_attrs)$/, keys(%$attr)): keys(%$attr);

	return ($open . join(' ', uc($node->{'tag'}),
	  map(format_attribute($_, $attr->{$_}), sort @keys)) . $close);
}

1;
