#
# Render::HTML -- Render in plain HTML
#
# Copyright (c) 2002, 2003 Brian Katzung, Kappa Computer Solutions, LLC.
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2002-09-25	Rewritten by Brian Katzung.
# 2002-09-17	Created by Brian Katzung.
#

package Render::HTML;

use strict;
use Parse::XMLLite qw(dispatch);

#----------------------------------------------------------------------
# Append formatting codes (public instance method)

sub append
{
	my ($self, $text) = @_;

	push(@{$self->{'text'}}, $text);
}

#----------------------------------------------------------------------
# Format tree (private instance method)

sub _format
{
	my ($self, $node) = @_;

#print "_format type $node->{'type'} tag ", $node->{'tag'} || '', "\n";

	return if (dispatch($node,
	'b-tag' =>
	{
		'body' => sub { $self->_format_BODY($node); },
		'docinfo' => sub { $self->_format_DOCINFO($node); },
		'em' => sub { $self->_format_EM($node); },
		'head' => sub { $self->_format_HEAD($node); },
		'table' => sub { $self->_format_TABLE($node); },
	},
	'e-tag' =>
	{
		'body' => sub { ; },
		'em' => sub { ; },
		'head' => sub { ; },
		'table' => sub { $self->_format_TABLE($node); },
		'ul' => sub { $self->_format_end_list($node); },
	},
	'tag' =>
	{
		'docinfo' => sub { $self->_format_DOCINFO($node); },
		'pagenum' => sub { ; },
	},
	));

	print {$self->{'file'}} $self->_recode($node);
	$self->{'did_table'} = 0 unless ($node->{'type'} eq 'space');
	$self->_format_content($node);
}

#----------------------------------------------------------------------
# Format the BODY node (private instance method)

sub _format_BODY
{
	my ($self, $node) = @_;

	$self->_format_HEAD($node);

	print {$self->{'file'}} <<EOT;
</BODY>
</HTML>
EOT
}

#----------------------------------------------------------------------
# Format node content (private instance method)

sub _format_content
{
	my ($self, $node) = @_;

	foreach (@{$node->{'content'} || []})
	{
		$self->_format($_);
	}
}

#----------------------------------------------------------------------
# Format the DOCINFO node (private instance method)

sub _format_DOCINFO
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};

	$self->{'title'} = $attr->{'title'} if ($attr->{'title'});
}

#----------------------------------------------------------------------
# Format the EM (emphasis) node (private instance method)

sub _format_EM
{
	my ($self, $node) = @_;

	print {$self->{'file'}} "<STRONG>";
	$self->_format_content($node);
	print {$self->{'file'}} "</STRONG>";
}

#----------------------------------------------------------------------
# Format an end-list node (hack for IE formatting quirks)
# (private instance method)

sub _format_end_list
{
	my ($self, $node) = @_;
	my $next = $node->{'parent'}{'next'};

	while ($next && $next->{'type'} =~ /^(?:comment|e-tag|space)$/)
	{
		$next = $next->{'next'};
	}

#my $pp = $node->{'parent'}{'parent'};
#print "_f_e_l parent $pp->{'type'}/$pp->{'tag'}\n";
	print {$self->{'file'}} $self->_recode($node)
	  if ($next || $node->{'parent'}{'parent'}{'tag'} !~ /^t[dh]$/);
}

#----------------------------------------------------------------------
# Format the HEAD node (private instance method)

sub _format_HEAD
{
	my ($self, $node) = @_;

	$self->_render_head();
	$self->_format_content($node);
}

#----------------------------------------------------------------------
# Format an option value (private function)

sub _format_option
{
	my ($key, $value) = @_;

	$value =~ s/</&lt;/g;

	return "$key='$value'" unless ($value =~ /'/);
	return "$key=\"$value\"" unless ($value =~ /"/);

	$value =~ s/"/&quot;/g;

	return "$key=\"$value\"";
}

#----------------------------------------------------------------------
# Format a TABLE node (private instance method)

sub _format_TABLE
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'} || {};

	if ($node->{'type'} eq 'e-tag')
	{
		print {$self->{'file'}} "</TABLE>\n";
		++$self->{'did_table'};
		return;
	}

	if ($self->{'did_table'})
	{
		print {$self->{'file'}} $self->{'table_spacer'};
		$self->{'did_table'} = 0;
	}

	print {$self->{'file'}} $self->_recode(
	{
		'type' => 'b-tag',
		'tag' => 'table',
		'attribute' =>
		{
			'border' => 0,
			'cellpadding' => 0,
			'cellspacing' => 0,
			%$attr
		},
	}), "\n";

	$self->_format_content($node);
}

#----------------------------------------------------------------------
# Return a new drawing object (public class method)

sub new
{
	my ($class, $file) = @_;
	my $self = bless
	{
		'file' => $file || *STDOUT,
		'table_spacer' => "<P></P>\n",
		'text' => [],
		'title' => 'Untitled',
	}, $class;

	return $self;
}

#----------------------------------------------------------------------
# Recode a node (private instance method)

sub _recode
{
	my ($self, $node) = @_;
	my $type = $node->{'type'};

	return ($node->{'text'} =~ /\n/? "\n": ' ') if ($type eq 'space');
	if ($type eq 'text')
	{
		my $text = $node->{'text'};

		$text =~ s/&xtsep;/&nbsp;&nbsp;&nbsp;&nbsp;/g;
		return $text;
	}

	my $attr = $node->{'attribute'};
	my $open = $type eq 'e-tag'? '</': '<';
	my @keys = sort(keys(%$attr));

	my $result = ($open . join(' ', uc($node->{'tag'}),
	  map(_format_option($_, $attr->{$_}), @keys)) . '>');
#print "recode result = $result\n";
	return $result;
}

#----------------------------------------------------------------------
# Render formatting codes now (public instance method)

sub render
{
	my ($self) = @_;
	my $parser = Parse::XMLLite->new();
	my $root;

	$root = $self->{'root'} =
	  $parser->parse(join('', @{$self->{'text'}}));
	%$self =
	(
		%$self,
		'did_head' => 0,
		'did_table' => 0,
		'text' => [],
	);
	$self->_format_content($root);
}

#----------------------------------------------------------------------
# Generate the HTML HEAD section (private instance method)

sub _render_head
{
	my ($self) = @_;

	return if ($self->{'did_head'});

	print {$self->{'file'}} <<EOT;
<HTML>
<HEAD>
<TITLE>$self->{'title'}</TITLE>
</HEAD>
<BODY bgcolor='#FFFFFF'>
EOT

	++$self->{'did_head'};
}

1;
