#
# Render::Text -- Render as plain text
#
# Copyright (c) 2002 Brian Katzung, Kappa Computer Solutions, LLC.
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2002-09-26	Created by Brian Katzung.
#

package Render::Text;

use strict;
use vars qw(@bullets @entity_map);

use Render::TextCanvas;
use Render::TextTable;
use Parse::XMLLite qw(dispatch);

@bullets = qw(* > - o);

@entity_map = (
	[ '&#46;', '.' ],
	[ '&#64;', '@' ],
	[ '&#150;', '-' ],
	[ '&#8211;', '-' ],
	[ '&#146;', "'" ],
	[ '&#8217;', "'" ],
	[ '&middot;', '.' ],
	[ '&nbsp;', "\177" ],
	[ '&quot;', '"' ],
	[ '&xtsep;', "\177\177" ],
	[ '&amp;', '&' ],
);

#----------------------------------------------------------------------
# Align text onto a canvas (private function)

sub _align_text
{
	my ($align, $text, $canvas) = @_;
	my $width = $canvas->get_width();

	if ($align eq 'left')
	{
		$canvas->draw($text);
	}
	elsif ($align eq 'center')
	{
		foreach (@$text)
		{
			$canvas->draw(' ' x int(($width - length($_)) / 2) .
			  $_);
		}
	}
	else
	{
		foreach (@$text)
		{
			$canvas->draw(' ' x ($width - length($_)) .
			  $_);
		}
	}
}

#----------------------------------------------------------------------
# Append text to input code

sub append
{
	my ($self, $text) = @_;

	push(@{$self->{'input'}}, $text);
}

#----------------------------------------------------------------------
# Append text to the text list (private instance method)

sub _append_text
{
	my ($self, $canvas, @text) = @_;

	foreach (@text)
	{
		foreach my $map (@entity_map)
		{
			$_ =~ s/$map->[0]/$map->[1]/g;
			$_ = uc($_) if ($self->{'emphasis'});
		}
	}
	$self->{'talign'} = $self->{'force_left'}? 'left': $self->{'align'};
	$self->{'tcanvas'} = $canvas;
	push(@{$self->{'tlist'}}, @text);
}

#----------------------------------------------------------------------
# Set/get compact mode (public instance method)

sub compact
{
	my ($self, $com) = @_;
	my $old_com = $self->{'compact'};

	$self->{'compact'} = $com if (defined($com));
	return $old_com;
}

#----------------------------------------------------------------------
# Flush any accumulated text (private instance method)

sub _flush_text
{
	my ($self) = @_;
	my $canvas = $self->{'tcanvas'};
	my $text = join('', @{$self->{'tlist'}});
	my $min;

	$self->{'tlist'} = [];
	$text =~ s/\s+/ /g;
	$text =~ s/^ | $//g;

	return 0 if ($text eq '');

	($text, $min) = _wrap_text($text, $canvas->get_width());
	_align_text($self->{'talign'}, $text, $canvas);
	$canvas->min_width($min);

	return 1;
}

#----------------------------------------------------------------------
# Format a node (private instance method)

# B I
# BODY BR
# DOCINFO
# H1 H2 H3 H4 H5 H6
# HEAD
# P OL UL LI
# TABLE TD TR

sub _format
{
	my ($self, $node, $canvas) = @_;

	return if (dispatch($node,
	'b-tag' =>
	{
		'b' => sub { $self->_format_B($node, $canvas); },
		'docinfo' => sub { ; },
		'em' => sub { $self->_format_EM($node, $canvas); },
		'head' => sub
		{
			$self->_format_content($node, $canvas);
			$self->_format_BR($node, $canvas);
		},
		'h1' => sub { $self->_format_H12($node, $canvas); },
		'h2' => sub { $self->_format_H12($node, $canvas); },
		'h3' => sub { $self->_format_Hn($node, $canvas); },
		'h4' => sub { $self->_format_Hn($node, $canvas); },
		'h5' => sub { $self->_format_Hn($node, $canvas); },
		'h6' => sub { $self->_format_Hn($node, $canvas); },
		'i' => sub { $self->_format_I($node, $canvas); },
		'ol' => sub { $self->_format_list($node, $canvas); },
		'p' => sub { $self->_format_P($node, $canvas); },
		'table' => sub { $self->_format_TABLE($node, $canvas); },
		'ul' => sub { $self->_format_list($node, $canvas); },
	},
	'space' => sub { $self->_append_text($canvas, ' '); },
	'tag' =>
	{
		'br' => sub { $self->_format_BR($node, $canvas); },
		'docinfo' => sub { ; },
	},
	'text' => sub { $self->_format_text($node, $canvas); },
	));

	$self->_format_content($node, $canvas);
}

#----------------------------------------------------------------------
# Format a B (bold) node (private instance method)

sub _format_B
{
	my ($self, $node, $canvas) = @_;

	$self->_append_text($canvas, '*');
	$self->_format_content($node, $canvas);
	$self->_append_text($canvas, '*');
}

#----------------------------------------------------------------------
# Format a BR (break) node (private instance method)

sub _format_BR
{
	my ($self, $node, $canvas) = @_;

	$canvas->draw('') unless ($self->_flush_text());
}

#----------------------------------------------------------------------
# Format node content (private instance method)

sub _format_content
{
	my ($self, $node, $canvas) = @_;

	foreach (@{$node->{'content'} || []})
	{
		$self->_format($_, $canvas);
	}
}

#----------------------------------------------------------------------
# Format an EM (emphasis) node (private instance method)

sub _format_EM
{
	my ($self, $node, $canvas) = @_;
	my $old_em = $self->{'emphasis'};

	$self->{'emphasis'} = 1;
	$self->_format_content($node, $canvas);
	$self->{'emphasis'} = $old_em;
}

#----------------------------------------------------------------------
# Format an H1 or H2 node (private instance method)

sub _format_H12
{
	my ($self, $node, $canvas) = @_;

	$self->_flush_text();
	$canvas->draw('') unless ($self->_is_first($node));

	if ($self->{'compact'})
	{
		#
		# ## H1 ##
		# + H2 +
		#
		my $marker = $node->{'tag'} eq 'h1'? '##': '+';

		$self->_append_text($canvas, "$marker&nbsp;");
		$self->_format_content($node, $canvas);
		$self->_append_text($canvas, "&nbsp;$marker");
		$self->_flush_text();
	}
	else
	{
		#
		# H1
		# ==
		#
		# H2
		# --
		#
		my $h_canvas = Render::TextCanvas->new($canvas->get_width());
		my $hchr = $node->{'tag'} eq 'h1'? '=': '-';

		$self->_format_content($node, $h_canvas);
		$self->_flush_text();
		$canvas->min_width($h_canvas->min_width());
		$h_canvas->draw($hchr x ($h_canvas->get_size())[0]);
		_align_text($self->{'force_left'}? 'left': $self->{'align'},
		  scalar($h_canvas->get_lines()), $canvas);
	}
}

#----------------------------------------------------------------------
# Format an Hn (for n > 2) node (private instance method)

sub _format_Hn
{
	my ($self, $node, $canvas) = @_;

	$self->_flush_text();
	$canvas->draw('') unless ($self->_is_first($node));
	$self->_format_content($node, $canvas);
	$self->_flush_text();
}

#----------------------------------------------------------------------
# Format an I (italic) node (private instance method)

sub _format_I
{
	my ($self, $node, $canvas) = @_;

	$self->_append_text($canvas, '/');
	$self->_format_content($node, $canvas);
	$self->_append_text($canvas, '/');
}

#----------------------------------------------------------------------
# Format an LI node (private instance method)

sub _format_LI
{
	my ($self, $node, $canvas) = @_;
	my $parent = $node->{'parent'};

	return unless ($parent->{'type'} eq 'b-tag' &&
	  ($parent->{'tag'} eq 'ol' || $parent->{'tag'} eq 'ul'));

	my $label = ($parent->{'tag'} eq 'ul')? $parent->{'bullet'}:
	  $node->{'parent'}{'next_num'}++ . '.';

	if ($self->{'compact'} && $self->{'compact'} > 1)
	{
		$self->_flush_text();
		$self->_append_text($canvas, "$label ");
		$self->_format_content($node, $canvas);
		$self->_flush_text();
		return;
	}

	my $tr_node =
	{
		'type' => 'b-tag',
		'tag' => 'tr',
		'attribute' =>
		{
			'valign' => 'top',
		},
		'content' =>
		[
			{
				'type' => 'b-tag',
				'tag' => 'td',
				'attribute' =>
				{
					'align' => 'right',
				},
				'content' =>
				[
					{
						'type' => 'text',
						'text' => "$label&nbsp;",
					},
				],
			},
			{
				'type' => 'b-tag',
				'tag' => 'td',
				'attribute' =>
				{
					'align' => 'left',
				},
				'content' => $node->{'content'},
			},
		],
		'parent' => $parent->{'list_table'},
	};
	$tr_node->{'content'}[0]{'parent'} = $tr_node;
	$tr_node->{'content'}[1]{'parent'} = $tr_node;
	foreach (@{$node->{'content'} || []})
	{
		$_->{'parent'} = $tr_node->{'content'}[1];
	}
	push(@{$parent->{'list_table'}{'content'}}, $tr_node);
}

#----------------------------------------------------------------------
# Format a list (OL or UL) node (private instance method)

sub _format_list
{
	my ($self, $node, $canvas) = @_;

	$self->_flush_text();
	$canvas->draw('') unless ($self->_is_first($node));
	$self->{'list_level'} ||= 0;
	$node->{'next_num'} = 1;
	++$self->{'list_level'};
	$node->{'bullet'} = $bullets[($self->{'list_level'} - 1) % @bullets];

	$node->{'list_table'} =
	{
		'type' => 'b-tag',
		'tag' => 'table',
#'attribute' =>
#{
#	'tborder' => 3,
#},
		'content' => [],
	}
	  unless ($self->{'compact'} && $self->{'compact'} > 1);

	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' => { 'li' => sub { $self->_format_LI($_, $canvas); }, },
		);
	}

	unless ($self->{'compact'} && $self->{'compact'} > 1)
	{
#$self->_append_text($canvas, 'list table here');
#$self->_flush_text();
		$self->_format($node->{'list_table'}, $canvas);
		delete $node->{'list_table'};
	}

	--$self->{'list_level'};
}

#----------------------------------------------------------------------
# Format a P (paragraph) node (private instance method)

sub _format_P
{
	my ($self, $node, $canvas) = @_;

	$self->_flush_text();
	$canvas->draw('') unless ($self->_is_first($node));
	$self->_format_content($node, $canvas);
	$self->_format_BR($node, $canvas);
}

#----------------------------------------------------------------------
# Accumulate raw text (private instance method)

sub _format_text
{
	my ($self, $node, $canvas) = @_;
	my $text = $node->{'text'};

	$self->_append_text($canvas, $text);
}

#----------------------------------------------------------------------
# Return true if node is the first node in the parent container
# (private instance method)

sub _is_first
{
	my ($self, $node) = @_;

	return 1 unless $node->{'parent'};

	foreach (@{$node->{'parent'}{'content'} || []})
	{
		return 1 if ($_ == $node);
		return 0 unless ($_->{'type'} =~ /^(?:comment|space)$/);
	}

	return 0;
}

#----------------------------------------------------------------------
# Create a new renderer instance (public class method)

sub new
{
	my ($class) = @_;
	my $self = bless
	{
		'align' => 'left',		# Current default alignment
		'input' => [],			# Raw input
		'tlist' => [],			# Text accumulation
		'width' => 70,			# Output width (line length)
	}, $class;

	return $self;
}

#----------------------------------------------------------------------
# Begin rendering (public instance method)

sub render
{
	my ($self) = @_;
	my $parser = Parse::XMLLite->new();
	my $root = $self->{'root'} =
	  $parser->parse(join('', @{$self->{'input'}}));
	my $canvas = $self->{'canvas'} =
	  Render::TextCanvas->new($self->{'width'});

	$self->{'align'} = 'left';
	$self->{'input'} = [];

	foreach (@{$root->{'content'} || []})
	{
		$self->_format($_, $canvas);
		print map("$_\n", $canvas->get_lines());
	}
}

#----------------------------------------------------------------------
# Set/get width (line length) (public instance method)

sub width
{
	my ($self, $width) = @_;
	my $old_width = $self->{'width'};

	$self->{'width'} = $width
	  if (defined($width) && $width >= 10);
	return $old_width;
}

#----------------------------------------------------------------------
# Wrap text to a specific line length (private function)

sub _wrap_text
{
	my ($text, $len) = @_;
	my @in = grep($_ ne '',
	  split(/( |[^ -]*-?\177*)/i, $text));
	my @current;
	my @out;
	my $min_len = 0;			# Desired min line length
	my $remaining = $len;
	my $space = 0;

unless ($len) {
print "Wrap to len ", ($len || 'undef'), ": >>$text<<\n";
exit 1;
}
	while (@in)
	{
		my $next = shift(@in);		# The next word or space

		while ($next =~ /\177$/ && @in)
		{
			$next .= shift(@in);
		}

		if ($next eq ' ')
		{
			#
			# Space: just remember that we saw it.
			#
			$space = 1;
			next;
		}

		my $next_len = length($next);

		#
		# Keep a running tally of the longest word.
		#
		$min_len = $next_len if ($next_len > $min_len);

		if ($next_len > $len)
		{
			#
			# The word is too long to fit on a single line at
			# this length. Just put as much as possible onto
			# each line.
			#
			my $part =
			  substr($next, 0, $remaining - $space, '');

			if (@current)
			{
				#
				# A partial line is already in progress.
				#
				if ($part ne '')
				{
					push(@current, ' ') if ($space);
					push(@current, $part);
				}
				push(@out, join('', @current));
				@current = ();
				$remaining = $len;
				$space = 0;
			}
			else
			{
				#
				# The entire line width is available for
				# this chunk.
				#
				push(@out, $part);
			}

			#
			# Push back the remainder of the word for the next
			# iteration.
			#
			unshift(@in, $next);
			next;
		}

		if (@current && $next_len + $space > $remaining)
		{
			push(@out, join('', @current));
			@current = ();
			$remaining = $len;
			$space = 0;
		}

		if ($space)
		{
			push(@current, ' ');
			$space = 0;
			--$remaining;
		}
		push(@current, $next);
		$remaining -= $next_len;
	}

	push(@out, join('', @current)) if (@current);

	foreach (@out)
	{
		s/\177/ /g;
	}

	return (\@out, $min_len);
}

1;
