#
# Render::TextCanvas -- A simple text canvas
#
# Copyright (c) 2002 Brian Katzung, Kappa Computer Solutions, LLC.
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2002-09-26	Created by Brian Katzung
#

package Render::TextCanvas;

use strict;

#----------------------------------------------------------------------
# Append text lines to the bottom of the canvas (public instance method)

sub draw
{
	my $self = shift(@_);
	my $lines = ref($_[0])? $_[0]: \@_;

	$self->draw_at(0, $self->{'Yn'}, $lines);
}

#----------------------------------------------------------------------
# Draw text lines somewhere on the canvas (public instance method)

sub draw_at
{
	my ($self, $x, $y) = splice(@_, 0, 3);
	my $lines = ref($_[0])? $_[0]: \@_;
	my $max_len = $self->{'width'} - $x;
	my @lines = ($max_len <= 0)? ('') x @$lines:
	  map(substr($_, 0, $max_len), @$lines);

	if ($x < 0)
	{
		@lines = map(substr($_, -$x), @lines);
		$x = 0;
	}

	foreach (@lines)
	{
		$self->_raw_draw($x, $y++, $_);
	}
}

#----------------------------------------------------------------------
# Read and remove lines from the canvas (public instance method)

sub get_lines
{
	my ($self) = @_;
	my $lines = $self->read(0, $self->{'Yt'},
	  $self->{'width_used'}, $self->{'Yn'} - $self->{'Yt'});

	$self->{'canvas'} = [] if ($self->{'canvas'});
	$self->{'Yt'} = $self->{'Yn'};
	$self->{'min_width'} = $self->{'width_used'} = 0;

	return wantarray? @$lines: $lines;
}

#----------------------------------------------------------------------
# Return how much canvas has been used as (width, current height,
# total height) (public instance method)

sub get_size
{
	my ($self) = @_;

	return ($self->{'width_used'}, $self->{'Yn'} - $self->{'Yt'},
	  $self->{'Yn'});
}

#----------------------------------------------------------------------
# Return canvas width (public instance method)

sub get_width
{
	my ($self) = @_;

	return $self->{'width'};
}

#----------------------------------------------------------------------
# Set/get the desired minimum width (public instance method)
# (This is primarily used for sizing table cells and has no impact
# on the canvas.)

sub min_width
{
	my ($self, $min) = @_;
	my $old_min = $self->{'min_width'};

	if ($min && $min > $self->{'min_width'})
	{
		$self->{'min_width'} = $min;
		$self->{'parent'}->min_width($min) if ($self->{'parent'});
	}
	return $old_min;
}

#----------------------------------------------------------------------
# Create a new instance (public class method)

sub new
{
	my ($class, $width, $parent, $Xo, $Yo) = @_;
	my $self = bless
	{
		'min_width' => 0,
		'width_used' => 0,
		'Yn' => 0,			# Next Y in viewport
		'Yt' => 0,			# Top Y in viewport
	}, $class;

	if (defined($parent))
	{
		#
		# Viewport onto parent canvas or viewport
		#
		$Xo ||= 0;
		$Yo = $parent->{'Yn'} unless (defined($Yo));

		my $max_width = $parent->{'width'} - $Xo;

		$width = $max_width if ($width > $max_width);
		$width = 0 if ($width < 0);
		%$self = (%$self,
			'parent' => $parent,
			'width' => $width,
			'Xo' => $Xo,		# X offset in parent
			'Yo' => $Yo,		# Y offset in parent
		);
	}
	else
	{
		#
		# Canvas
		#
		$width = 0 if ($width < 0);
		%$self = (%$self,
			'canvas' => [],		# Text data
			'width' => $width,
		);
	}

	return $self;
}

#----------------------------------------------------------------------
# Draw a pre-trimmed line at a particular spot on the canvas
# (private instance method)

sub _raw_draw
{
	my ($self, $x, $y, $line) = @_;
	my $width = length($line);
	my $parent = $self->{'parent'};

	if ($parent)
	{
		#
		# Viewport: pass to parent canvas for drawing.
		#
		$parent->_raw_draw($self->{'Xo'} + $x,
		  $self->{'Yo'} + $y, $line);
	}
	elsif ($width)
	{
		#
		# Extend the canvas if necessary and insert the text.
		#
		my $canvas = $self->{'canvas'};
		my $index = $y - $self->{'Yt'};

		if ($index >= 0)
		{
			$canvas->[$index] = ' ' x $self->{'width'}
			  unless (defined($canvas->[$index]));
			substr($canvas->[$index], $x, $width) = $line;
		}
	}

	#
	# Update stats from the perspective of the canvas or
	# viewport as appropriate.
	#
	$width += $x;
	$self->{'width_used'} = $width
	  if ($width > $self->{'width_used'});
	$self->{'Yn'} = $y + 1 if ($y >= $self->{'Yn'});
}

#----------------------------------------------------------------------
# Read text from the canvas (public instance method)

sub read
{
	my ($self, $Xo, $Yo, $width, $height) = @_;
	my $view = $self;
	my @text;

	$width = $self->{'width_used'} if ($width > $self->{'width_used'});
	$height = $self->{'Yn'} if ($height > $self->{'Yn'});

	while ($view->{'parent'})
	{
		$Xo += $view->{'Xo'};
		$width = $view->{'width'} - $Xo
		  if ($Xo + $width > $view->{'width'});
		$Yo += $view->{'Yo'};
		$view = $view->{'parent'};
	}

	$Yo -= $view->{'Yt'};
	while (--$height >= 0)
	{
		my $text = $view->{'canvas'}[$Yo++];

		$text = '' unless (defined($text));
		$text = substr(' ' x -$Xo . $text, $Xo, $width);
		$text =~ s/\s+$//;
		push(@text, $text);
	}

	return wantarray? @text: \@text;
}

1;
