#
# Render::TextTable -- Table processing for text rendering
#
# Copyright (c) 2002 Brian Katzung, Kappa Computer Solutions, LLC.
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2002-09-27	Created by Brian Katzung
#

package Render::Text;

use strict;
use vars qw(@border_chars);

#
# Default border definitions:
#
# +---+  *****  #####
# |   |  *   *  #   #
# | 1 |  * 2 *  # 3 #
# |   |  *   *  #   #
# +---+  *****  #####
#

@border_chars = (
	# [ corner, horizontal, vertical ]
	[ ' ', ' ', ' ' ],
	[ '+', '-', '|' ],
	[ '*', '*', '*' ],
	[ '#', '#', '#' ],
);

#----------------------------------------------------------------------
# Assemble borders, padding, spacing, and cell content onto the
# table canvas (private instance method)

sub _assemble_TABLE
{
	my ($self, $table, $t_canvas) = @_;
	my $attr = $table->{'attribute'};
	my $border = $attr->{'tborder'} || 0;
	my $any_border = $border? 1: 0;
	my ($cchr, $hchr, $vchr) =
	  @{$border_chars[$border] || $border_chars[$#border_chars]};
	my $pad = $attr->{'tcellpadding'} || 0;
	my $space = $attr->{'tcellspacing'} || 0;
	my $fmt = $table->{'fmt'};
	my $cell = $fmt->{'cell'};
	my $ncols = $fmt->{'ncols'};
	my $nrows = $fmt->{'nrows'};
	my $maxy = $cell->[$nrows - 1][0]{'row'}{'btm'} + $any_border + $space;
	my @corners;

	if ($border)
	{
		my $rpt = $fmt->{'outer_width'} - 2;

		$t_canvas->draw($cchr . ($hchr x $rpt) . $cchr);
		$t_canvas->draw(($vchr . (' ' x $rpt) . $vchr) x ($maxy - 2));
		$t_canvas->draw($cchr . ($hchr x $rpt) . $cchr);
	}
	else
	{
		$t_canvas->draw(('') x $maxy);
	}

	for (my $y = 0; $y < $nrows; ++$y)
	{
		my $Xo = $space;
		my $Yo = $cell->[$y][0]{'row'}{'top'};
		my $next_x;

		for (my $x = $next_x = 0; $x < $ncols; $x = $next_x)
		{
			my $cxy = $cell->[$y][$x];
			my $c_canvas = $cxy->{'canvas'};
			my $width = $cell->[0][$x]{'col'}{'use'};

			unless ($c_canvas)
			{
				$Xo += $width;
				++$next_x;
				next;
			}

			my $node = $cxy->{'node'};
			my $height =
			  $cell->[$node->{'end_row'}][0]{'row'}{'btm'} - $Yo;
			my $colspan = $cxy->{'node'}{'attribute'}{'colspan'}
			  || 1;

			#
			# Compute total width of cell, including spanned
			# columns.
			#
			$next_x = $x + $colspan;
			while ($colspan > 1)
			{
				$width +=
				  $cell->[0][$x + --$colspan]{'col'}{'use'};
			}

			#
			# Draw the border first, if applicable.
			#
			if ($border)
			{
				my $rpt = $width - $space - 1;
				my $b_canvas = Render::TextCanvas->new(
				  $rpt + 2, $t_canvas, $Xo, $Yo);

				$b_canvas->draw($cchr . ($hchr x $rpt));
				$b_canvas->draw(($vchr . (' ' x $rpt)
				  . $vchr) x ($height - 1));
				$b_canvas->draw($cchr . ($hchr x $rpt));

				push(@corners, [ $Xo + $rpt + 1, $Yo ],
				  [ $Xo, $Yo + $height ],
				  [ $Xo + $rpt + 1, $Yo + $height]);
			}

			#
			# Now copy the content.
			#
			my $c_height = ($c_canvas->get_size())[1];
			my $valign = lc($node->{'attribute'}{'valign'} ||
			  $node->{'parent'}{'attribute'}{'valign'} ||
			  'top');

#print "Rendering cell $x, $y at $Xo, $Yo\n";
			if ($valign eq 'bottom')
			{
				$t_canvas->draw_at($Xo + $any_border + $pad,
				  $Yo + $height - $c_height + $pad,
				  $c_canvas->get_lines());
			}
			elsif ($valign eq 'middle')
			{
				$t_canvas->draw_at($Xo + $any_border + $pad,
				  $Yo + int(($height - $c_height) / 2) +
				  $any_border + $pad, $c_canvas->get_lines());
			}
			else # top
			{
				$t_canvas->draw_at($Xo + $any_border + $pad,
				  $Yo + $any_border + $pad,
				  $c_canvas->get_lines());
			}

			$Xo += $width;
		}
	}

	#
	# Fill in any right-hand border corners. These are done last
	# so that left-hand borders of cells to the right don't clobber
	# corner characters.
	#
	foreach (@corners)
	{
		$t_canvas->draw_at(@$_, $cchr);
	}
}

#----------------------------------------------------------------------
# Format a TABLE node (private instance method)

sub _format_TABLE
{
	my ($self, $node, $canvas) = @_;
	my $attr = $node->{'attribute'};
	my $cwidth = $canvas->get_width();
	my $fmt = $node->{'fmt'} = {};

	$self->_flush_text();
	$canvas->draw('') unless ($self->_is_first($node));

	#
	# First guess at outer table width:
	# 1) Absolute value in text width attribute
	# 2) Width of trial canvas
	# 3) % of trial canvas in text width attribute
	# This may get adjusted in _set_column_widths.
	#
	my $trialw = $attr->{'twidth'} || $cwidth;

	$trialw = int($cwidth * substr($trialw, 0, -1) / 100)
	  if ($trialw =~ /%$/);

	#
	# Initialize table layout data and trial-render the content.
	#
	$fmt->{'outer_width'} = $trialw;
	$fmt->{'cell'} = [];		# The cell grid
	$fmt->{'ncols'} = 0;		# # of columns in the table
	$fmt->{'nrows'} = 0;		# # of rows in the table

	#
	# Initialize rows
	#
	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' => { 'tr' => sub { $self->_init_TR($_); }, },
		);
	}

	unless ($fmt->{'nrows'} && $fmt->{'ncols'})
	{
#print "Table has no extent! (rows $fmt->{'nrows'} cols $fmt->{'ncols'})\n";
		delete $node->{'fmt'};
		return;
	}

	#
	# Rendering pass
	#
	# First, settle on exact column widths and the table outer width.
	#
	_set_column_widths($node, $canvas);

	my $align = $self->{'force_left'}? 'left':
	  lc($attr->{'align'} || $self->{'align'});
	my $twidth = $fmt->{'outer_width'};
	my $table_canvas;

	if ($align eq 'center')
	{
		$table_canvas = Render::TextCanvas->new($twidth, $canvas,
		  int(($cwidth - $twidth) / 2));
	}
	elsif ($align eq 'right')
	{
		$table_canvas = Render::TextCanvas->new($twidth, $canvas,
		  $cwidth - $twidth);
	}
	else # left
	{
		$table_canvas = Render::TextCanvas->new($twidth, $canvas);
	}

	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' => { 'tr' => sub { $self->_format_TR($_); }, },
		);
	}

	_set_row_heights($node);
	$self->_assemble_TABLE($node, $table_canvas);

	#
	# Free up the memory containing the bulk of the layout details.
	#
	delete $node->{'fmt'};
}

#----------------------------------------------------------------------
# Format a TD (table cell) node (private instance method)

sub _format_TD
{
	my ($self, $node) = @_;
	my $row_node = $node->{'parent'};
	my $table = $row_node->{'parent'};
	my $t_attr = $table->{'attribute'};
	my $border = $t_attr->{'tborder'}? 1: 0;
	my $pad = $t_attr->{'tpadding'} || 0;
	my $space = $t_attr->{'tspacing'} || 0;
	my $fmt = $table->{'fmt'};
	my $col = $node->{'col'};
	my $colspan = $node->{'attribute'}{'colspan'} || 1;
	my $width = $fmt->{'cell'}[0][$col]{'col'}{'use'};

	while ($colspan > 1)
	{
		$width +=
		  $fmt->{'cell'}[0][$col + --$colspan]{'col'}{'use'};
	}

	my $canvas = $fmt->{'cell'}[$node->{'row'}][$col]{'canvas'} =
	  Render::TextCanvas->new($width - $border - $space - $pad * 2);
	my $save_align = $self->{'align'};

	$self->{'align'} = $node->{'attribute'}{'align'} ||
	  (($node->{'tag'} eq 'th')? 'center':
	  ($row_node->{'attribute'}{'align'} || 'left'));
	$self->_format_content($node, $canvas);
	$self->_flush_text();
	$self->{'align'} = $save_align;
}

#----------------------------------------------------------------------
# Format a TR (table row) node (private instance method)

sub _format_TR
{
	my ($self, $node) = @_;
	my $table = $node->{'parent'};

	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' =>
		{
			'td' => sub { $self->_format_TD($_); },
			'th' => sub { $self->_format_TD($_); },
		},
		);
	}
}

#----------------------------------------------------------------------
# Initialize a TD (table cell) node (private instance method)

sub _init_TD
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $row_node = $node->{'parent'};
	my $table = $row_node->{'parent'};
	my $fmt = $table->{'fmt'};
	my $row_num = $node->{'row'} = $row_node->{'row'};
	my $col_num = $node->{'col'} = _next_column($table, $row_num);
	my $colspan = $attr->{'colspan'} || 1;
	my $rowspan = $attr->{'rowspan'} || 1;
	my $end_row = $node->{'end_row'} = $row_num + $rowspan - 1;
	my $cellw = $attr->{'twidth'};
	my ($cminw, $min_per);
	my ($cmaxw, $max_per);
	my ($csize, $size_per) = 0;
	my ($pct, $pct_per) = 0;

	if ($cellw && $cellw !~ /%$/)
	{
		#
		# This is a fixed-width cell.
		#
		$cminw = $cmaxw = $cellw;
	}
	else
	{
		#
		# Trial-format the content as a basis for auto-sizing.
		#
		my $trialw = $fmt->{'outer_width'} *
		  substr($cellw || '100%', 0, -1) / 100;
		my $save_force = $self->{'force_left'};
		my $canvas = Render::TextCanvas->new($trialw);

		#
		# Format the content left-aligned for the trial rendering.
		# We don't want the cell profile skewed by centering or
		# right-alignment space.
		#
		$self->{'force_left'} = 1;
		$self->_format_content($node, $canvas);
		$self->_flush_text();
		$self->{'force_left'} = $save_force;

		#
		# Note the minimum cell width needed to not have to split
		# any words during wrapping and the maximum width to which
		# the cell should expand in order to reduce line wrapping
		# (the estimated width of the table being an upper bound).
		#
		$cminw = $canvas->min_width();
		$cmaxw = $colspan > 1? $cminw: ($canvas->get_size())[0];
		$csize = length(join('', $canvas->get_lines()));
		if ($cellw && $cellw =~ /%$/)
		{
			$pct = substr($cellw, 0, -1);
			$fmt->{'has_pct_cols'} = 1;
		}
	}

	#
	# Enter the cell into the cell grid, distributing sizing data
	# across the spanned columns as evenly as possible and
	# distributing the node pointer across both row and column spans.
	#
	my $align = $attr->{'align'} ||
	  $row_node->{'attribute'}{'align'} ||
	  ($node->{'tag'} eq 'td'? 'left': 'center');
	my @col_nums = ($col_num .. $col_num + $colspan - 1);

	if ($align eq 'right')
	{
		#
		# Favor later columns because the content will be
		# right-aligned by default.
		#
		@col_nums = reverse (@col_nums);
	}
	elsif ($align eq 'center')
	{
		#
		# Favor middle columns because the content will be
		# centered by default.
		#
		my @new;

		while (@col_nums)
		{
			push(@new, shift(@col_nums));
			push(@new, pop(@col_nums)) if (@col_nums);
		}
		@col_nums = @new;
	}

#print "Cell min $cminw max $cmaxw size $csize $pct%\n";
	$min_per = int(($cminw + $colspan - 1) / $colspan);
	$max_per = int(($cmaxw + $colspan - 1) / $colspan);
	$size_per = int(($csize + $colspan - 1) / $colspan);
	$pct_per = int(($pct + $colspan - 1) / $colspan);

	$fmt->{'ncols'} = $col_nums[-1] + 1
	  if ($col_nums[-1] >= $fmt->{'ncols'});

	my $save_size = $csize;
	my $save_pct = $pct;

	foreach my $x (@col_nums)
	{
		my $colp = ($fmt->{'cell'}[0][$x]{'col'} ||=
		{
			'max' => 0,
			'min' => 0,
			'size' => 0,
		});

		$min_per = $cminw if ($cminw < $min_per);
		$max_per = $cmaxw if ($cmaxw < $max_per);
		$size_per = $csize if ($csize < $size_per);
		$pct_per = $pct if ($pct < $pct_per);

		$colp->{'max'} = $max_per if ($max_per > $colp->{'max'});
		$colp->{'min'} = $min_per if ($min_per > $colp->{'min'});
		$colp->{'pct'} = $pct_per
		  if ($save_pct && $pct_per > ($colp->{'pct'} || 0));
		$colp->{'size'} += $size_per;

		$cminw -= $min_per;
		$cmaxw -= $max_per;
		$csize -= $size_per;
		$pct -= $pct_per;

		foreach my $y ($row_num .. $end_row)
		{
			$fmt->{'cell'}[$y][$x]{'node'} = $node;
		}
	}
}

#----------------------------------------------------------------------
# Initialize a TR (table row) node (private instance method)

sub _init_TR
{
	my ($self, $node) = @_;
	my $table = $node->{'parent'};

	$node->{'row'} = $table->{'fmt'}{'nrows'}++;

	#
	# Initialize cells
	#
	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' =>
		{
			'td' => sub { $self->_init_TD($_); },
			'th' => sub { $self->_init_TD($_); },
		},
		);
	}
}

#----------------------------------------------------------------------
# Return the next available column in a table cell grid (private function)

sub _next_column
{
	my ($table, $row) = @_;
	my $cell = $table->{'fmt'}{'cell'}[$row] || [];
	my $i;

	for ($i = 0; $i < @$cell; ++$i)
	{
		return $i unless (defined($cell->[$i]));
	}

	return $i;
}

#----------------------------------------------------------------------
# Set table column widths (private function)

sub _set_column_widths
{
	my ($table, $canvas) = @_;
	my $attr = $table->{'attribute'};
	my $fmt = $table->{'fmt'};
	my $cell = $fmt->{'cell'};
	my $texpf = 0;			# Table expansion factor
	my $tmaxw = 0;			# Table max width
	my $tminw = 0;			# Table min width

	#
	# 1. Foreach column, compute the minimum width needed (max over all
	# cells in the column), the maximum desired width (max over all
	# cells in the column), and the combined size of all auto-sizing
	# cells in the column. [Handled in _init_TD.]
	#
	# 2. For columns containing %-width cells, the %-width of the column
	# is the max of %-width for all %-width cells in the column.
	# [Handled in _init_TD.]
	#
	# 3. Compute the minimum width PER percent (max over all %-width
	# columns) and the average desired maximum width PER percent across
	# across all columns.
	#
	# 4. Minimum width and maximum desired width for columns containing
	# %-width cells are then recomputed from the aggregate characteristics.
	#
	# 5. Compute the minimum required width and desired maximum width for
	# the entire table (excluding borders/padding/spacing) by summing
	# across all columns. For each column, compute an "expansion factor",
	# the product of expansion (max width - min width) and size.
	#
	# 6. Compute the width available to cells from the specified or
	# potential maximum table size and border/padding/spacing options.
	#
	# 7. If the minimum width is >= the available space, assign the minimum
	# width to each column (thus "oversizing" the table) and stop.
	#
	# 8. If the maximum desired width is <= the available space and the
	# table is auto-sizing, assign the maximum desired width to each
	# column and stop.
	#
	# 9. The available expansion width is the available width less the
	# table minimum width. Allocate the available expansion to columns
	# in proportion to the expansion factor.
	#

	if ($fmt->{'has_pct_cols'})
	{
		my $tminpp = 0;
		my $tavgpp = 0;
		my $tpct = 0;

		#
		# Step 3 -- Min & avg width per percent.
		#
		for (my $x = $fmt->{'ncols'}; --$x >= 0; )
		{
			my $colp = $fmt->{'cell'}[0][$x]{'col'};
			my $pct;

			if ($pct = $colp->{'pct'})
			{
				my $cminpp = int($colp->{'min'} * 1000 / $pct);

				$tminpp = $cminpp if ($cminpp > $tminpp);
				$tavgpp += $colp->{'max'};
				$tpct += $pct;
			}
		}

		$tavgpp = int($tavgpp * 1000 / $tpct);
		$tavgpp = $tminpp if ($tavgpp < $tminpp);

		#
		# Step 4 -- Recompute min & max for % columns.
		#
		for (my $x = $fmt->{'ncols'}; --$x >= 0; )
		{
			my $colp = $fmt->{'cell'}[0][$x]{'col'};
			my $pct;

			if ($pct = $colp->{'pct'})
			{
				$colp->{'min'} = int(($tminpp * $pct + 999)
				  / 1000);
				$colp->{'max'} = int(($tavgpp * $pct + 999)
				  / 1000);
			}
		}
	}

	#
	# Step 5 -- Compute table min and max and column expansion factors.
	#
	for (my $x = $fmt->{'ncols'}; --$x >= 0; )
	{
		my $colp = $cell->[0][$x]{'col'};

		$tmaxw += $colp->{'max'};
		$tminw += $colp->{'min'};
		$texpf +=
		  ($colp->{'expf'} = ($colp->{'max'} - $colp->{'min'}) *
		  $colp->{'size'});
	}

	#
	# Step 6 -- Compute available inner table width.
	#
	my $border = $attr->{'tborder'}? 1: 0;
	my $pad = $attr->{'tcellpadding'} || 0;
	my $space = $attr->{'tcellspacing'} || 0;
	my $twidth = $attr->{'twidth'};

	if ($attr->{'twidth'})
	{
		if ($attr->{'twidth'} =~ /%$/)
		{
			$fmt->{'outer_width'} = int($canvas->get_width() *
			  substr($twidth, 0, -1) / 100);
		}
		else
		{
			$fmt->{'outer_width'} = $twidth;
		}
	}
	else
	{
		$fmt->{'outer_width'} = $canvas->get_width();
	}

	my $bps =
	  ($border + $space) * ($fmt->{'ncols'} + 1) +
	  $pad * $fmt->{'ncols'} * 2;

	my $avail = $fmt->{'outer_width'} - $bps;
#print "outer $fmt->{'outer_width'} - bps $bps = avail $avail\n";

	#
	# Step 7 -- If there isn't any expansion capacity, go with the
	# column minimums, even if it "oversizes" the table (in which
	# case only part of the table will appear on the canvas).
	#
	if ($avail <= $tminw)
	{
		for (my $x = $fmt->{'ncols'}; --$x >= 0; )
		{
			$cell->[0][$x]{'col'}{'use'} =
			  $cell->[0][$x]{'col'}{'min'} +
			  $border + $space + $pad * 2;
		}

		$fmt->{'outer_width'} = $tminw + $bps;
#print "exact/oversize min $tminw + bps $bps = outer $fmt->{'outer_width'}\n";
		return;
	}

	#
	# Step 8 -- Use the maximums as is if possible and the table
	# is auto-sizing.
	#
	if ($tmaxw <= $avail && !$twidth)
	{
		for (my $x = $fmt->{'ncols'}; --$x >= 0; )
		{
			$cell->[0][$x]{'col'}{'use'} =
			  $cell->[0][$x]{'col'}{'max'} +
			  $border + $space + $pad * 2;
		}

		$fmt->{'outer_width'} = $tmaxw + $bps;
#print "auto max $tmaxw + bps $bps = outer $fmt->{'outer_width'}\n";
		return;
	}

	#
	# Step 9 -- Allocate the available expansion in proportion to
	# column expansion factors.
	#
	my $inner = 0;
	$avail -= $tminw;
	for (my $x = $fmt->{'ncols'}; --$x >= 0; )
	{
		my $colp = $cell->[0][$x]{'col'};
		my $alloc = $texpf? int($avail * $colp->{'expf'} / $texpf): 0;

		$inner += ($colp->{'use'} = $colp->{'min'} + $alloc);
		$colp->{'use'} += $border + $space + $pad * 2;
		$avail -= $alloc;
		$texpf -= $colp->{'expf'};
	}

	if ($avail && $twidth)
	{
		for (my $x = $fmt->{'ncols'}; $tmaxw && --$x >= 0; )
		{
			my $colp = $cell->[0][$x]{'col'};
			my $alloc = int($avail * $colp->{'max'} / $tmaxw);

			$colp->{'use'} += $alloc;
			$inner += $alloc;
			$avail -= $alloc;
			$tmaxw -= $colp->{'max'};
		}
	}
	$fmt->{'outer_width'} = $inner + $bps;
#print "auto inner $inner + bps $bps = outer $fmt->{'outer_width'}\n";
}

#----------------------------------------------------------------------
# Set row heights (private function)

sub _set_row_heights
{
	my ($table) = @_;
	my $attr = $table->{'attribute'};
	my $border = $attr->{'tborder'}? 1: 0;
	my $pad = $attr->{'tcellpadding'} || 0;
	my $space = $attr->{'tcellspacing'} || 0;
	my $fmt = $table->{'fmt'};
	my $cell = $fmt->{'cell'};

	for (my $row = 0; $row < $fmt->{'nrows'}; ++$row)
	{
		$cell->[$row][0]{'row'}{'top'} =
		  ($row? $cell->[$row - 1][0]{'row'}{'btm'}: 0) +
		  $space;

		for (my $col = 0; $col < $fmt->{'ncols'}; ++$col)
		{
			my $canvas = $cell->[$row][$col]{'canvas'};

			next unless ($canvas);

			my $bottom = $cell->[$row][0]{'row'}{'top'} +
			  ($canvas->get_size())[1] + $border + $pad * 2;
			my $end_row = $cell->[$row][$col]{'node'}{'end_row'};

			$cell->[$end_row][0]{'row'}{'btm'} = $bottom
			  if ($bottom > ($cell->[$end_row][0]{'row'}{'btm'}
			  || 0));
		}
	}
}

1;
