#
# Resume -- Resume formatting object
#
# Copyright (c) 2002, 2003 Brian Katzung, Kappa Computer Solutions, LLC.
# All rights reserved.
#
#   This program is free software WHEN USED FOR NON-COMMERCIAL PURPOSES;
#   you can redistribute it and/or modify it under the terms of the GNU
#   General Public License as published by the Free Software Foundation;
#   either version 2 of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information or commercial-use licensing, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2002-09-17	Created by Brian Katzung
#

package Resume;

use strict;
use Parse::XMLLite qw(dispatch format_attribute);

sub FORMAT_BLOCK { 1; }
sub FORMAT_INLINE { 2; }

#----------------------------------------------------------------------
# Append a new node after an existing node (private instance method)

sub _append_node
{
	my ($self, $current, $new) = @_;

	$new->{'prev'} = $current;
	$new->{'next'} = $current->{'next'};
	$new->{'parent'} = $current->{'parent'};
	$current->{'next'} = $new;
	$self->_fix_parent_content($current);
}

#----------------------------------------------------------------------
# Set/get excerpt string

sub excerpts
{
	my ($self, $str) = @_;
	my $old_str = $self->{'excerpts'};

	$self->{'excerpts'} = $str if (defined($str));
	return $old_str;
}

#----------------------------------------------------------------------
# Filter sub-tree content for priority and target (private instance method)
# Returns true if the sub-tree should be retained

sub _filter_content
{
	my ($self) = @_;

	unless ($self->{'filtered'})
	{
		$self->_filter_1_content($self->{'root'});
		$self->{'filtered'} = 1;
	}
}

sub _filter_1_content
{
	my ($self, $node) = @_;
	my @content = @{$node->{'content'} ||= []};

	#
	# Always retain nodes with no content.
	#
	return 1 unless (@content);

	my $pri = $self->{'priority'};
	my $target = $self->{'target'};
	my @order = ([]);
	my $has_content = 0;
	my $excerpted = 0;
	my $space;

	#
	# Scan the contents.
	#
	while (@content)
	{
		my $next = shift(@content);

		#
		# Strip comments.
		#
		next if ($next->{'type'} eq 'comment');

		if ($next->{'type'} eq 'space')
		{
			#
			# Just remember the last space.
			#
			$space = $next;
			next;
		}

		if ($next->{'type'} eq 'e-tag')
		{
			$space = $next;
			last;
		}

		#
		# Retain "in-lines" where the target controls the presence
		# or absence of the effect rather than inclusion or exclusion
		# of content.
		#
		if ($next->{'type'} eq 'b-tag' &&
		  $next->{'tag'} =~ /^(?:b|em|i)$/)
		{
			push(@{$order[0]}, $space) if ($space);
			push(@{$order[0]}, $next);
			$space = undef;
			$has_content = 1;
			next;
		}

		#
		# Filter out low-priority content if a threshold is set.
		#
		if ($pri)
		{
			my $n_pri = ',' .
			  ($next->{'attribute'}{'priority'} || 0) .  ',';

			($n_pri) = ($n_pri =~ /,(?:$target:)?(\d+),/);
			if ($n_pri && $n_pri > $pri)
			{
				$excerpted = 1;
				$space = undef;
				next;
			}
		}

		#
		# Get the next target. Content without a target is
		# always included.
		#
		my $n_target = ',' .
		  ($next->{'attribute'}{'target'} || '*') .
		  ',everything,';

		#
		# Target selection:
		# An explicitly included target is always included.
		# An explicitly excluded target is excluded (unless
		# also explicitly included!).
		# Exclusion is implied if there is a wildcard exclusion.
		# Inclusion is implied if there is a wildcard inclusion
		# or any target is an exclusion.
		# Otherwise, exclusion is implied.
		#
		if ($n_target =~ /,$target(?::(\d*)|,)/ ||
		  ($n_target !~ /,!(?:$target|\*),/ &&
		  ($n_target =~ /,\*(?::(\d*)|,)/ || $n_target =~ /,!/)))
		{
			my $rank = $1 || 0;

			if ($next->{'type'} ne 'b-tag' ||
			  $self->_filter_1_content($next))
			{
				$order[$rank] ||= [];
				push(@{$order[$rank]}, $space) if ($space);
				push(@{$order[$rank]}, $next);
				$has_content = 1;
			}
		}

		$space = undef;
	}

	push(@{$order[0]}, $space) if ($space);

	#
	# Rebuild content in rank order (plus "other").
	#
	my $other = shift(@order);
	foreach (@order)
	{
		push(@content, @$_) if ($_);
	}
	push(@content, @$other);

	$node->{'content'} = \@content;
	$self->_relink_content($node);		# Fix prev/next links

	#
	# Allow customization of attributes with <SET/>.
	#
	foreach (@content)
	{
		next if ($_->{'type'} eq 'space');

		#
		# Allow attributes of first <SET/> to override original
		# attributes except 'target'.
		#
		$node->{'attribute'} = { %{$node->{'attribute'}},
		  %{$_->{'attribute'}},
		  'target' => $node->{'attribute'}{'target'} }
		  if ($_->{'type'} eq 'tag' && $_->{'tag'} eq 'set');

		#
		# Stop after the first non-space content regardless.
		#
		last;
	}

	#
	# Track nodes with excerpted (sub-)content.
	#
	$node->{'parent'}{'excerpted'} ||= $node->{'excerpted'} ||= $excerpted;

	#
	# Retain this node if any content remains, or if marked to
	# always be retained.
	#
	return ($has_content || $node->{'attribute'}{'retain'});
}

#----------------------------------------------------------------------
# Find a node (and its parent) matching the criteria supplied as a closure
# (public instance method)

sub find_node
{
	my ($self, $matcher, $node) = @_;

	return Parse::XMLLite::find_node($matcher, $node || $self->{'root'});
}

#----------------------------------------------------------------------
# Fix parent content array after inserting or appending a node
# (private instance method)

sub _fix_parent_content
{
	my ($self, $node) = @_;
	my @content = ($node);		# New content is the "current" node...

	# ... plus all the ones that come before...
	for (my $prev = $node->{'prev'}; $prev; $prev = $prev->{'prev'})
	{
		unshift(@content, $prev);
	}

	# ... plus all the ones that come after.
	for (my $next = $node->{'next'}; $next; $next = $next->{'next'})
	{
		push(@content, $next);
	}

	$node->{'parent'}{'content'} = \@content;
}

#----------------------------------------------------------------------
# Format and render the (first) RESUME (public instance method)

sub format
{
	my ($self) = @_;

	#
	# Filter and order content based on target and priority.
	#
	$self->_filter_content();

	#
	# Find and format the first RESUME.
	#
	my $resume = $self->find_node(sub
	{
		my ($node) = @_;

		return ($node->{'type'} eq 'b-tag' &&
		  $node->{'tag'} eq 'resume');
	});

	if ($resume)
	{
		%$self = (%$self,
			'clients' => [],	# Client stack
			'emps' => [],		# Employer stack
			'resume' => $resume,	# Resume node
			's_level' => 0,		# Section level
		);
		$self->_format_RESUME($resume);
	}
}

#----------------------------------------------------------------------
# Format a CL (column list) node (private instance method)

sub _format_CL
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};

	my $table = { %$node, 'tag' => 'table' };
	my $cols = $node->{'attribute'}{'columns'} || 2;
	my $li = 0;

	delete $table->{'attribute'}{'columns'};
	foreach (@{$node->{'content'}})
	{
		next if ($_->{'type'} ne 'b-tag' || $_->{'tag'} ne 'li');

		#
		# Start the table on the first item.
		#
		$self->{'draw'}->append($self->_recode($table))
		  if ($li == 0);

		if ($li % $cols == 0)
		{
			#
			# Start a new row before each first column.
			#
			$self->{'draw'}->append("<TR valign='top'>");
		}
		else
		{
			#
			# Add an empty column between real columns.
			#
			$self->{'draw'}->append('<TD>&nbsp;</TD>');
		}

		#
		# Put in the real list item.
		#
		$self->{'draw'}->append('<TD><UL>');
		$self->_format_LI($_);
		$self->{'draw'}->append('</UL></TD>');

		#
		# Finish the row after each last column.
		#
		$self->{'draw'}->append("</TR>\n")
		  if (++$li % $cols == 0);
	}

	#
	# End the table if started.
	#
	$self->{'draw'}->append("</TABLE>\n") if ($li != 0);
}

#----------------------------------------------------------------------
# Format a CLIENT node (private instance method)

sub _format_CLIENT
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $name = $attr->{'name'} || '';

	push(@{$self->{'clients'}}, $node)
	  unless ($name eq '');

	#
	# Check for immediate header display.
	#
	if ($attr->{'display'})
	{
		$self->_format_work_header();
		if ($name eq '')
		{
			$node->{'parent'}{'did_header'} = 1;
		}
		else
		{
			$node->{'did_header'} = 1;
		}
	}

	#
	# Process any content.
	#
	foreach (@{$node->{'content'} || []})
	{
		my $type = $_->{'type'};

		next unless ($type eq 'b-tag' || $type eq 'tag');

		#
		# Content that doesn't force a header display.
		#
		next if (dispatch($_,
		'b-tag' =>
		{
			'client' => sub { $self->_format_CLIENT($_); },
		},
		'tag' =>
		{
			'client' => sub { $self->_format_CLIENT($_); },
		},
		));

		#
		# All other content requires prior display of header.
		#
		unless ($node->{'did_header'})
		{
			$self->_format_work_header();
			$node->{'did_header'} = 1;
		}

		dispatch($_,
		'b-tag' =>
		{
			'cl' => sub { $self->_format_CL($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'section' => sub { $self->_format_SECTION($_); },
			'span' =>
			  sub { $self->_format_SPAN($_, FORMAT_BLOCK); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		},
		);
	}

	pop(@{$self->{'clients'}})
	  unless ($name eq '');
}

#----------------------------------------------------------------------
# Format a degree node (private instance method)

sub _format_DEGREE
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $began = $attr->{'began'};
	my $ended = $attr->{'completed'} || $self->{'dict'}{'Present'}
	  || 'Present';
	my $text =
	  "$attr->{'degree'}, $attr->{'study'}; $attr->{'location'}; ";

	$text .= $began . ($self->{'dict'}{'date_sep'} || '&nbsp;-&nbsp;')
	  if ($began);
	$text .= $ended if ($ended);

	$self->{'draw'}->append($text);
}

#----------------------------------------------------------------------
# Format a B, EM, or I (emphasis) node (private instance method)

sub _format_EM
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $target = $self->{'target'};
	my $n_target = ',' . ($attr->{'target'} || '*') . ',everything,';

	#
	# Emphasize, possibly conditional on target.
	#
	if ($n_target =~ /,$target,/ ||
	  ($n_target !~ /,!(?:$target|\*),/ &&
	  ($n_target =~ /,\*,/ || $n_target =~ /,!/)))
	{
		$self->{'draw'}->append("<\U$node->{'tag'}>");
		foreach (@{$node->{'content'}})
		{
			$self->_format_inline($_);
		}
		$self->{'draw'}->append("</\U$node->{'tag'}>");
	}
	else
	{
		foreach (@{$node->{'content'}})
		{
			$self->_format_inline($_);
		}
	}
}

#----------------------------------------------------------------------
# Format an EMPLOYER node (private instance method)

sub _format_EMPLOYER
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $name = $attr->{'name'} || '';

	#
	# Remember the heading level for nested CLIENT and
	# EMPLOYER tags.
	#
	unless ($name eq '')
	{
		$self->{'e_level'} = ++$self->{'s_level'}
		  if (@{$self->{'emps'}} == 0);
		push(@{$self->{'emps'}}, $node);
		$self->{'sel_emps'} = 0;
	}

	#
	# Check for selection of number of EMPLOYERs to display.
	#
	my $save_sel = $self->{'sel_emps'};
	$self->{'sel_emps'} = $attr->{'select'}
	  if ($attr->{'select'});

	#
	# Check for immediate header display.
	#
	if ($attr->{'display'})
	{
		$self->_format_work_header();
		if ($name eq '')
		{
			$node->{'parent'}{'did_header'} = 1;
		}
		else
		{
			$node->{'did_header'} = 1;
		}
	}

	#
	# Process content.
	#
	foreach (@{$node->{'content'} || []})
	{
		my $type = $_->{'type'};

		next unless ($type eq 'b-tag' || $type eq 'tag');

		#
		# Content that doesn't force a header display.
		#
		next if (dispatch($_,
		'b-tag' =>
		{
			'client' => sub { $self->_format_CLIENT($_); },
			'employer' => sub { $self->_format_EMPLOYER($_); },
		},
		'tag' =>
		{
			'client' => sub { $self->_format_CLIENT($_); },
			'employer' => sub { $self->_format_EMPLOYER($_); },
		},
		));

		#
		# All other content requires prior display of header.
		#
		unless ($node->{'did_header'})
		{
			$self->_format_work_header();
			$node->{'did_header'} = 1;
		}

		dispatch($_,
		'b-tag' =>
		{
			'cl' => sub { $self->_format_CL($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'section' => sub { $self->_format_SECTION($_); },
			'span' =>
			  sub { $self->_format_SPAN($_, FORMAT_BLOCK); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		},
		);
	}

	unless ($name eq '')
	{
		pop(@{$self->{'emps'}});
		--$self->{'s_level'}
		  if (@{$self->{'emps'}} == 0);
	}

	$self->{'sel_emps'} = $save_sel
	  if ($node->{'type'} eq 'b-tag');
}

#----------------------------------------------------------------------
# Format in-line content (private instance method)

sub _format_inline
{
	my ($self, $node) = @_;

	dispatch($node,
	'b-tag' =>
	{
		'b' => sub { $self->_format_EM($node); },
		'em' => sub { $self->_format_EM($node); },
		'i' => sub { $self->_format_EM($node); },
		'skill' => sub { $self->_format_SKILL($node); },
		'span' => sub { $self->_format_SPAN($node, FORMAT_INLINE); },
	},
	'cdata' => sub { $self->{'draw'}->append($node->{'text'}); },
	'space' => sub { $self->{'draw'}->append(' '); },
	'tag' =>
	{
		'br' => sub { $self->{'draw'}->append($self->_recode($node)); },
		'degree' => sub { $self->_format_DEGREE($node); },
	},
	'text' => sub { $self->{'draw'}->append($node->{'text'}); },
	);
}

#----------------------------------------------------------------------
# Format an LI (list item) node (private instance method)

sub _format_LI
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};

	$self->{'draw'}->append($self->_recode($node));
	foreach (@{$node->{'content'}})
	{
		#
		# Block content...
		#
		next if (dispatch($_,
		'b-tag' =>
		{
			'cl' => sub { $self->_format_CL($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'span' => sub { $self->_format_SPAN($_,
			  FORMAT_BLOCK | FORMAT_INLINE); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		},
		));

		#
		# In-line content...
		#
		$self->_format_inline($_);
	}
	$self->{'draw'}->append("</LI>\n");
}

#----------------------------------------------------------------------
# Format an OL or UL list node (private instance method)

sub _format_list
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};

	$self->{'draw'}->append($self->_recode($node));
	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' => { 'li' => sub { $self->_format_LI($_); } });
	}
	$self->{'draw'}->append($node->{'tag'} eq 'ol'? "</OL>\n": "</UL>\n");
}

#----------------------------------------------------------------------
# Format a P (paragraph) node (private instance method)

sub _format_P
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $label = $attr->{'label'};

	delete $attr->{'label'};
	$self->{'draw'}->append($self->_recode($node));

	#
	# Label this paragraph?
	#
	$self->{'draw'}->append("<B>$label</B> ") if (defined($label));

	#
	# Process content.
	#
	foreach (@{$node->{'content'}})
	{
		$self->_format_inline($_);
	}
	$self->{'draw'}->append("</P>\n");
}

#----------------------------------------------------------------------
# Format a RESUME node (private instance method)

sub _format_RESUME
{
	my ($self, $resume) = @_;
	my $attr = $resume->{'attribute'};

	$self->_preamble();
	foreach (@{$resume->{'content'}})
	{
		dispatch($_,
		'b-tag' =>
		{
			'section' => sub { $self->_format_SECTION($_); },
		},
		);
	}
	$self->_postamble();
}

#----------------------------------------------------------------------
# Format a SECTION node (private instance method)

sub _format_SECTION
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $label = $attr->{'label'};

	#
	# Handle labeling for labeled sections.
	#
	if (defined($label))
	{
		my $level = ++$self->{'s_level'};

		#
		# Level 1 may include "(Excerpts)" modifier.
		#
		if ($level == 1 && $self->{'priority'} &&
		  $self->{'excerpts'} ne '' && $node->{'excerpted'})
		{
			$self->_header(1, "$label $self->{'excerpts'}");
		}
		else
		{
			$self->_header($level, $label);
		}
	}

	if (($attr->{'type'} || '') eq 'skills' &&
	  $self->_tabulate_skills($node))
	{
		--$self->{'s_level'} if (defined($label));
		return;
	}

	#
	# Process content.
	#
	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' =>
		{
			'cl' => sub { $self->_format_CL($_); },
			'employer' => sub { $self->_format_EMPLOYER($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'section' => sub { $self->_format_SECTION($_); },
			'span' =>
			  sub { $self->_format_SPAN($_, FORMAT_BLOCK); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		},
		);
	}

	--$self->{'s_level'} if (defined($label));
}

#----------------------------------------------------------------------
# Format a SKILL node (private instance method)

sub _format_SKILL
{
	my ($self, $node) = @_;

	return unless ($node->{'content'});

	foreach (@{$node->{'content'}})
	{
		$self->_format_inline($_);
	}

	my $next = $node->{'next'};

	while ($next)
	{
		my $type = $next->{'type'};

		if ($type eq 'space' ||
		  ($type eq 'tag' && $next->{'tag'} eq 'skill'))
		{
			$next = $next->{'next'};
			next;
		}

		$self->{'draw'}->append('; ')
		  if ($type eq 'b-tag' && $next->{'tag'} eq 'skill');

		last;
	}
}

#----------------------------------------------------------------------
# Format a SPAN node (private instance method)

sub _format_SPAN
{
	my ($self, $node, $format) = @_;

	$self->{'draw'}->append($self->_recode($node));
	foreach (@{$node->{'content'}})
	{
		next if (($format & FORMAT_BLOCK) &&
		  dispatch($_,
		  'b-tag' =>
		  {
			'cl' => sub { $self->_format_CL($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		  },
		  ));

		$self->_format_inline($_)
		  if ($format & FORMAT_INLINE);
	}
	$self->{'draw'}->append("</SPAN>\n");
}

#----------------------------------------------------------------------
# Format a TABLE node (private instance method)

sub _format_TABLE
{
	my ($self, $node) = @_;

	$self->{'draw'}->append($self->_recode($node));
	foreach (@{$node->{'content'}})
	{
		dispatch($_,
		'b-tag' => { 'tr' => sub { $self->_format_TR($_); } });
	}
	$self->{'draw'}->append("</TABLE>\n");
}

#----------------------------------------------------------------------
# Format a TD/TH (table data) node (private instance method)

sub _format_TD
{
	my ($self, $node) = @_;

	$self->{'draw'}->append($self->_recode($node));
	foreach (@{$node->{'content'}})
	{
		next if (dispatch($_,
		'b-tag' =>
		{
			'cl' => sub { $self->_format_CL($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'section' => sub { $self->_format_SECTION($_); },
			'span' => sub { $self->_format_SPAN($_,
			  FORMAT_BLOCK | FORMAT_INLINE); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		},
		));

		$self->_format_inline($_);
	}
	$self->{'draw'}->append("</\U$node->{'tag'}>\n");
}

#----------------------------------------------------------------------
# Format a TR (table row) node (private instance method)

sub _format_TR
{
	my ($self, $node) = @_;
	my $fmt = sub { $self->_format_TD($_); };

	$self->{'draw'}->append($self->_recode($node));
	foreach (@{$node->{'content'}})
	{
		dispatch($_, 'b-tag' => { 'td' => $fmt, 'th' => $fmt });
	}
	$self->{'draw'}->append("</TR>\n");
}

#----------------------------------------------------------------------
# Format a work (employer or client) header (private instance method)

sub _format_work_header
{
	my ($self) = @_;
	my $draw = $self->{'draw'};

	#
	# Examine only EMPLOYERs and CLIENTs with information to print.
	#
	my @sel_emps =
	  grep($_->{'attribute'}{'name'} &&
	  ($_->{'attribute'}{'began'} || $_->{'attribute'}{'duration'}),
	  @{$self->{'emps'}});
	my @clients =
	  grep($_->{'attribute'}{'name'} &&
	  ($_->{'attribute'}{'began'} || $_->{'attribute'}{'duration'}),
	  @{$self->{'clients'}});

	#
	# Work with subset of EMPLOYERs if a "select" attribute is
	# in effect.
	#
	@sel_emps = splice(@sel_emps, -$self->{'sel_emps'})
	  if ($self->{'sel_emps'});

	return unless (@sel_emps);

	#
	# Get the work header format, with possible temporary local override.
	#
	my $fmt = $sel_emps[-1]->{'work_fmt'};
	if ($fmt)
	{
		my $old_fmt = $self->{'work_fmt'};

		$self->work_format($fmt);
		$fmt = $self->{'work_fmt'};
		$self->{'work_fmt'} = $old_fmt;
	}
	else
	{
		$fmt = $self->{'work_fmt'};
	}

	#
	# In partial format, format just the end client as if it were
	# a sole employer.
	#
	if (@clients && $fmt !~ /\*/)
	{
		@sel_emps = pop(@clients);
		@clients = ();
	}

	my $sep = $fmt =~ /;/? '; ': '<BR/>';

	my $date_sep = $self->{'dict'}{'date_sep'} || '&nbsp;-&nbsp;';
	my $present = $self->{'dict'}{'Present'} || 'Present';

	if (@clients)
	{
# Client format is:
# | title; employer [and employer] [for client] at client[, location] | began - ended |
# or
# | employer [and employer] [for client] at client[, location]; title | began - ended |
		my $client = pop(@clients);
		my $attr = $client->{'attribute'};
		my $ended = defined($attr->{'ended'})? $attr->{'ended'}:
		  $present;
		my $date;
		my $text = '';

		if ($attr->{'duration'})
		{
			$date = $attr->{'duration'};
		}
		else
		{
			$date = $attr->{'began'};
			$date .= "$date_sep$ended" unless ($ended eq $date);
		}

		$text .= $self->_wh_cond_em($client, 'title') . $sep
		  if ($attr->{'title'} && $fmt !~ /n/);

		my $emp = join(' ' . ($self->{'dict'}{'and'} || 'and') . ' ',
		  map($self->_wh_cond_em($_, 'name'), @sel_emps));

		$text .= join('', $emp,
		  map(' ' . ($self->{'dict'}{'for'} || 'for') . ' ' .
		  $self->_wh_cond_em($_, 'name'), @clients),
		  ' ' . ($self->{'dict'}{'at'} || 'at') . ' ' .
		  $self->_wh_cond_em($client, 'name'));

		$text .= ", $attr->{'location'}"
		  if ($attr->{'location'} && $fmt =~ /@/);

		$text .= $sep . $self->_wh_cond_em($client, 'title')
		  if ($attr->{'title'} && $fmt =~ /n/);

		$draw->append(
		  "<TABLE width='100%' twidth='100%'><TR valign='top'><TD>");
		if ($fmt =~ /h/)
		{
			$self->_header($self->{'e_level'}, $text);
		}
		else
		{
			$draw->append("<B>$text</B>");
		}
		$draw->append(
"</TD>\n<TD>&nbsp;&nbsp;&nbsp;</TD><TD align='right'>$date</TD></TR></TABLE>\n"
		  );
		return;
	}

	$draw->append("<TABLE width='100%' twidth='100%'>\n");

# Employer format is:
# | title; employer[, location] | began - ended |
# or
# | employer[, location]; title | began - ended |
	foreach (@sel_emps)
	{
		my $attr = $_->{'attribute'};
		my $ended = defined($attr->{'ended'})? $attr->{'ended'}:
		  $present;
		my $date;
		my $text = '';

		if ($attr->{'duration'})
		{
			$date = $attr->{'duration'};
		}
		else
		{
			$date = $attr->{'began'};
			$date .= "$date_sep$ended" unless ($ended eq $date);
		}

		$text .= $self->_wh_cond_em($_, 'title') . $sep
		  if ($attr->{'title'} && $fmt !~ /n/i);
		$text .= $self->_wh_cond_em($_, 'name');
		$text .= ", $attr->{'location'}"
		  if ($attr->{'location'} && $fmt =~ /@/);
		$text .= $sep . $self->_wh_cond_em($_, 'title')
		  if ($attr->{'title'} && $fmt =~ /n/i);

		$draw->append("<TR valign='top'><TD>");
		if ($fmt =~ /h/)
		{
			$self->_header($self->{'e_level'}, $text);
		}
		else
		{
			$draw->append("<B>$text</B>");
		}
		$draw->append(
"</TD>\n<TD>&nbsp;&nbsp;&nbsp;</TD><TD align='right'>$date</TD></TR>\n");
	}

	$draw->append("</TABLE>\n");
}

#----------------------------------------------------------------------
# Return the resume node (public instance method)

sub get_resume_node
{
	my ($self) = @_;

	return $self->{'resume'};
}

#----------------------------------------------------------------------
# Append a header (private instance method)

sub _header
{
	my ($self, $level, $label) = @_;

	$self->{'draw'}->append(
	  sprintf("<H%d>%s</H%d>\n", $level, $label, $level));
}

#----------------------------------------------------------------------
# Insert a new node before an existing node (private instance method)

sub _insert_node
{
	my ($self, $current, $new) = @_;

	$new->{'prev'} = $current->{'prev'};
	$new->{'next'} = $current;
	$new->{'parent'} = $current->{'parent'};
	$current->{'prev'} = $new;
	$self->_fix_parent_content($current);
}

#----------------------------------------------------------------------
# Create a new resume object from unparsed text (public class method)

sub new
{
	my ($class, $text) = @_;
	my $self = bless
	{
		'excerpts' => '(Excerpts)',	# Excerpt string or ''
		'priority' => 0,		# Default to everything
		'target' => 'default',		# Resume target
		# 1-line, incl loc, title 1st, partial cli info,
		# use headers
		'work_fmt' => ';@h',
		'dict' =>
		{
			'date_sep' => '&nbsp;- ',
		},
	}, $class;
	my $parser = Parse::XMLLite->new();

	$self->{'root'} = $parser->parse($text);

	return $self;
}

#----------------------------------------------------------------------
# Handle postamble (private instance method)

sub _postamble
{
	my ($self) = @_;

	$self->{'draw'}->append("</BODY>\n");
	$self->{'draw'}->render();
}

#----------------------------------------------------------------------
# Handle preamble (private instance method)

sub _preamble
{
	my ($self) = @_;
	my $resume = $self->get_resume_node();
	my $attr = $resume->{'attribute'};
	my $title = format_attribute('title',
	  ($self->{'dict'}{'Resume of'} || 'Resume of') . " $attr->{'name'}");
	my $updated = $self->{'dict'}{'Updated'} || 'Updated';
	my $phone = $self->{'dict'}{'Phone'} || 'Phone';
	my $fax = $self->{'dict'}{'FAX'} || 'FAX';
	my $email = $self->{'dict'}{'Email'} || 'Email';

	$self->{'draw'}->append(<<EOT);
<DOCINFO $title/>
<HEAD>
<TABLE width='100%' twidth='100%'><TR valign='top'>
<TD><H1>$attr->{'name'}</H1></TD>
<TD align='right'>$updated: $attr->{'updated'}<BR/>
<PAGENUM/></TD>
</TR></TABLE>
</HEAD>
<BODY>
<TABLE>
<TR><TD><TABLE>
<TR valign='top'><TD>$phone:</TD>
<TD rowspan='3'>&nbsp;&nbsp;</TD>
<TD>$attr->{'phone'}</TD></TR>
EOT
	$self->{'draw'}->append(<<EOT) if ($attr->{'fax'});
<TR valign='top'><TD>$fax:</TD>
<TD>$attr->{'fax'}</TD></TR>
EOT
	$self->{'draw'}->append(<<EOT) if ($attr->{'email'});
<TR valign='top'><TD>$email:</TD>
<TD><A href='mailto:$attr->{'email'}'>$attr->{'email'}</A></TD></TR>
EOT
	$self->{'draw'}->append(<<EOT);
</TABLE></TD>
<TD>&nbsp;&nbsp;&nbsp;&nbsp;</TD>
<TD>$attr->{'address'}<BR/>
EOT
	$self->{'draw'}->append("<A href='$attr->{'url'}'>$attr->{'url'}</A>")
	  if ($attr->{'url'});
	$self->{'draw'}->append('</TD></TR></TABLE>');
}

#----------------------------------------------------------------------
# Set/get priority level

sub priority
{
	my ($self, $pri) = @_;
	my $old_pri = $self->{'priority'};

	$self->{'priority'} = $pri if (defined($pri) && $pri >= 0);
	return $old_pri;
}

#----------------------------------------------------------------------
# Recode a node into text (private instance method)

sub _recode
{
	my ($self, $node) = @_;

	return Parse::XMLLite::unparse_node($node, 'priority|target');
}

#----------------------------------------------------------------------
# Relink content nodes (private instance method)

sub _relink_content
{
	my ($self, $node) = @_;
	my $prev;

	foreach (@{$node->{'content'} || []})
	{
		$_->{'parent'} = $node;
		if ($prev)
		{
			$_->{'prev'} = $prev;
			$prev->{'next'} = $_
		}
		$prev = $_;
	}
}

#----------------------------------------------------------------------
# Set resume rendering object (public instance method)

sub renderer
{
	my ($self, $renderer) = @_;

	return $self->{'draw'} = $renderer;
}

#----------------------------------------------------------------------
# Tabulate skills when appropriate (private instance method)
# Returns true if tabulated

sub _tabulate_skills
{
	my ($self, $node) = @_;
	my $tabulate = 0;

	#
	# Prescan for tabulatability.
	#
	$self->_ts_check_SECTION($node, \$tabulate);
	return 0 unless ($tabulate);

	#
	# Generate the skill table preamble.
	#
	my $draw = $self->{'draw'};
	my $skill = $self->{'dict'}{'Skill'} || 'Skill';
	my $first = $self->{'dict'}{'skill first'} || 'First<BR/>Used';
	my $last = $self->{'dict'}{'skill last'} || 'Last<BR/>Used';
	my $used = $self->{'dict'}{'skill used'} || 'Years<BR/>Used';
	my $level = $self->{'dict'}{'skill level'} || 'Level';

	$draw->append(<<EOT);
<TABLE width='100%' twidth='100%'>
<TR valign='bottom'><TH>$skill</TH><TD>&xtsep;</TD>
<TH>$first</TH><TD>&xtsep;</TD>
<TH>$last</TH><TD>&xtsep;</TD>
<TH>$used</TH><TD>&xtsep;</TD>
<TH align='left' width='25%' twidth='25%'>$level</TH>
<TD width='10%' twidth='10%'>&nbsp;</TD></TR>
<TR><TD>&nbsp;</TD></TR>
EOT

	#
	# Process the content.
	#
	$self->_ts_format_SECTION($_, 1);

	#
	# Finish the skill table.
	#
	$draw->append("</TABLE>\n");

	return 1;
}

#----------------------------------------------------------------------
# Set/get resume target (public instance method)

sub target
{
	my ($self, $target) = @_;
	my $old_target = $self->{'target'};

	$self->{'target'} = $target if (defined($target) && $target ne '');
	return $old_target;
}

#----------------------------------------------------------------------
# Check a P node for SKILL tabulatability (private instance method)

sub _ts_check_P
{
	my ($self, $node, $tabulate) = @_;

	#
	# Ignore this node if table attribute excludes it for
	# this target.
	#
	return unless ($self->_ts_check_table($node));

	foreach (@{$node->{'content'}})
	{
		my $attr = $_->{'attribute'};

		#
		# We want to tabulate if we find a skill with a name
		# (or text) and a level.
		#
		if (($_->{'tag'} || '') eq 'skill' &&
		  $self->_ts_check_table($_) && $attr->{'first'} &&
		  $attr->{'level'} && ($attr->{'name'} ||
		  $self->find_node(
		  sub { return $_[0]->{'type'} eq 'text' }, $_)))
		{
			$$tabulate = 1;
			return;
		}
	}
}

#----------------------------------------------------------------------
# Check a SECTION node for SKILL tabulatability (private instance method)

sub _ts_check_SECTION
{
	my ($self, $node, $tabulate) = @_;

	#
	# Ignore this node if table attribute excludes it for
	# this target.
	#
	return unless ($self->_ts_check_table($node));

	foreach (@{$node->{'content'}})
	{
		#
		# Scan P and SECTION nodes for tabulatable SKILLs.
		#
		last if (dispatch($_,
		'b-tag' =>
		{
			'p' => sub { $self->_ts_check_P($_, $tabulate); },
			'section' =>
			  sub { $self->_ts_check_SECTION($_, $tabulate); },
		},
		) && $$tabulate);
	}
}

#----------------------------------------------------------------------
# Return true if this node allows tabulation (private instance method)

sub _ts_check_table
{
	my ($self, $node) = @_;
	my $target = $self->{'target'};
	my $n_target = ',' . ($node->{'attribute'}{'table'} || '*') . ',';

	#
	# Target selection: See _filter_content.
	#
	return ($n_target =~ /,$target,/ ||
	  ($n_target !~ /,!(?:$target|\*),/ &&
	  ($n_target =~ /,\*,/ || $n_target =~ /,!/)));
}

#----------------------------------------------------------------------
# Format a P node with tabulated SKILLS (private instance method)

sub _ts_format_P
{
	my ($self, $node) = @_;
	my $draw = $self->{'draw'};
	my $this_year = (localtime())[5] + 1900;
	my @skills;
	my @content;

	#
	# Split content in tabulatable SKILLS and other.
	#
	while (@{$node->{'content'}})
	{
		my $next = shift(@{$node->{'content'}});
		my $attr = $next->{'attribute'};

		next if ($next->{'type'} eq 'e-tag');

		#
		# Tabulatable SKILLs get added to @skills.
		# Printable content gets added to @content (e.g. leading
		# space and non-printing SKILLs are excluded).
		#
		if (($next->{'tag'} || '') eq 'skill' &&
		  $self->_ts_check_table($next) &&
		  $attr->{'first'} && $attr->{'level'} &&
		  ($attr->{'name'} || $self->find_node(
		  sub { return $_[0]->{'type'} eq 'text'; })))
		{
			push(@skills, $next);
		}
		elsif ((@content || $next->{'type'} ne 'space') &&
		  ($next->{'type'} ne 'tag' || $next->{'tag'} ne 'skill'))
		{
			push(@content, $next);
		}
	}

	my $label = $node->{'attribute'}{'label'} || '';

	#
	# Place the paragraph label before table SKILLs, if any.
	#
	if ($label ne '' && @skills)
	{
		$draw->append("<TR><TD colspan='10'><B>$label</B></TD></TR>\n");
		delete $node->{'attribute'}{'label'};
	}

	#
	# Format tabulatable SKILLs into table rows.
	#
	foreach (@skills)
	{
		my $attr = $_->{'attribute'};
		my $skill = $attr->{'name'};
		my $first = $attr->{'first'};
		my $last = $attr->{'last'} || $this_year;
		my $used = $attr->{'used'} || ($last - $first + 1);
		my $level = $attr->{'level'};

		$draw->append(
		  "<TR align='right' valign='top'><TD align='left'>");
		if ($skill)
		{
			#
			# SKILL has a name attribute.
			#
			$draw->append($skill);
		}
		else
		{
			#
			# Use SKILL node content.
			#
			delete $_->{'prev'};
			delete $_->{'next'};
			$self->_format_inline($_);
		}

		#
		# Now add first/last/used/level.
		#
		$draw->append(<<EOT);
</TD><TD></TD>
<TD>$first</TD><TD></TD><TD>$last</TD><TD></TD>
<TD>$used</TD><TD></TD><TD align='left'>$level</TD></TR>
EOT
	}

	#
	# If there's any other content (including non-tabulatable SKILLs),
	# display that as a paragraph (wrapped in a table row) after the
	# SKILL rows.
	#
	if (@content)
	{
		$draw->append("<TR><TD>&nbsp;</TD></TR>\n") if (@skills);
		$draw->append("<TR><TD colspan='10'>");
		delete $node->{'prev'};
		delete $node->{'next'};
		$node->{'content'} = \@content;
		$self->_relink_content($node);

		$self->_format_P($node);

		$draw->append("</TD></TR>\n");
	}
}

#----------------------------------------------------------------------
# Format a SECTION node with tabulated SKILLS (private instance method)

sub _ts_format_SECTION
{
	my ($self, $node, $top) = @_;
	my $attr = $node->{'attribute'};
	my $label = $attr->{'label'};
	my $draw = $self->{'draw'};

	#
	# Handle labeling for labeled sections.
	#
	if (!$top && defined($label))
	{
		my $level = ++$self->{'s_level'};

		#
		# Wrap the header in a table row.
		#
		$draw->append("<TR><TD colspan='10'>");
		$self->_header($level, $label);
		$draw->append("</TD></TR><TR><TD>&nbsp;</TD></TR>\n");
	}

	#
	# Process content.
	#
	my $first = 1;
	foreach (@{$node->{'content'}})
	{
		next if ($_->{'type'} eq 'space' || $_->{'type'} eq 'e-tag');

		#
		# Add a spacing row between content.
		#
		if ($first)
		{
			$first = 0;
		}
		else
		{
			$draw->append("<TR><TD>&nbsp;</TD></TR>\n");
		}

		#
		# P and SECTION tags have custom formatting (unless
		# their table attribute excludes them).
		#
		next if ($self->_ts_check_table($_) && dispatch($_,
		'b-tag' =>
		{
			'p' => sub { $self->_ts_format_P($_); },
			'section' => sub { $self->_ts_format_SECTION($_); },
		},
		));

		#
		# Use normal formatting for most other SECTION content
		# but wrap it in a table row.
		#
		$draw->append("<TR><TD colspan='10'>");
		dispatch($_,
		'b-tag' =>
		{
			'cl' => sub { $self->_format_CL($_); },
			'ol' => sub { $self->_format_list($_); },
			'p' => sub { $self->_format_P($_); },
			'section' => sub { $self->_format_SECTION($_); },
			'span' =>
			  sub { $self->_format_SPAN($_, FORMAT_BLOCK); },
			'table' => sub { $self->_format_TABLE($_); },
			'ul' => sub { $self->_format_list($_); },
		},
		);
		$draw->append("</TD></TR>\n");
	}

	--$self->{'s_level'} if (!$top && defined($label));
}

#----------------------------------------------------------------------
# Return a work header attribute with conditional emphasis
# (private instance method)

sub _wh_cond_em
{
	my ($self, $node, $sel) = @_;
	my $marker = substr($sel, 0, 1);
	my $attr = $node->{'attribute'};
	my $val = $attr->{$sel};
	my $n_target = ',' . ($attr->{'target'} || '') . ',';
	my ($markers) = ($n_target =~ /,(?:$self->{'target'}|\*):([^,]*)/);

#print "$attr->{'name'}: $marker in ", $markers || '-', "\n";
	return (($markers && $markers =~ /$marker/)? "<EM>$val</EM>": $val);
}

#----------------------------------------------------------------------
# Set/get work header format (public instance method)

sub work_format
{
	my ($self, $fmt) = @_;
	my $old = $self->{'work_fmt'};

	if ($fmt)
	{
		my $tmp = $old;

		if ($fmt =~ /1/)	# 1-line
		{
			$tmp .= ';' unless ($tmp =~ /;/);
		}
		if ($fmt =~ '2')	# 2-line
		{
			$tmp =~ s/;//;
		}
		if ($fmt =~ /f/)	# full client info
		{
			$tmp .= '*' unless ($tmp =~ /\*/);
		}
		if ($fmt =~ /n/)	# name first
		{
			$tmp .= 'n' unless ($tmp =~ /n/);
		}
		if ($fmt =~ /p/)	# partial client info
		{
			$tmp =~ s/\*//;
		}
		if ($fmt =~ /t/)	# title first
		{
			$tmp =~ s/n//;
		}
		if ($fmt =~ /\+h/)	# use headers instead of bold
		{
			$tmp .= 'h' unless ($tmp =~ /h/);
		}
		if ($fmt =~ /-h/)	# don't use headers (use bold)
		{
			$tmp =~ s/h//;
		}
		if ($fmt =~ /\+l/)	# include location
		{
			$tmp .= '@' unless ($tmp =~ /\@/);
		}
		if ($fmt =~ /-l/)	# exclude location
		{
			$tmp =~ s/\@//;
		}

		$self->{'work_fmt'} = $tmp;
	}

	#
	# Map storage format to arg format for return value.
	#
	if ($old =~ /;/)
	{
		$old =~ s/;/1/;
	}
	else
	{
		$old .= '2';
	}

	if ($old =~ /\*/)
	{
		$old =~ s/\*/f/;
	}
	else
	{
		$old .= 'p';
	}

	$old .= 't' unless ($old =~ /n/);

	if ($old =~ /h/)
	{
		$old =~ s/h/+h/;
	}
	else
	{
		$old .= '-h';
	}

	if ($old =~ /\@/)
	{
		$old =~ s/\@/+l/;
	}
	else
	{
		$old .= '-l';
	}

	return $old;
}

1;
