#
# Resume::Functionalize -- Functionalize a resume object
#
# Copyright (c) 2003 Brian Katzung, Kappa Computer Solutions, LLC.
# All rights reserved.
#
#   This program is free software WHEN USED FOR NON-COMMERCIAL PURPOSES;
#   you can redistribute it and/or modify it under the terms of the GNU
#   General Public License as published by the Free Software Foundation;
#   either version 2 of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For more information or commercial-use licensing, contact:
#
#	Kappa Computer Solutions, LLC
#	3223 Lake Ave #15C-393
#	Wilmette, IL 60091-1074
#	877.367.8837 (US) or 781.370.1093
#	http://www.kappacs.com/
#
# 2003-04-03	Created by Brian Katzung
#
# Description:
# This module creates a functional-format resume from a chronological
# XML resume. This is done as follows:
#
# 1) The top-level sections are stripped from CLIENT and EMPLOYER
#    sections within a SECTION of type 'work'. The immediately enclosing
#    CLIENT or EMPLOYER tag is noted for each SECTION.
#
# 2) The remaining original section content (CLIENT tags, EMPLOYER tags,
#    and anything outside of a SECTION) becomes part of a new "work history"
#    section. This may be displayed before or after work experience if there
#    is an history='before' or history='after' attribute, respectively. When
#    present, an history_label attribute should also be provided to label
#    the new section.
#
# 3) The stripped client/employer SECTIONs are grouped by label. The content
#    of each group is combined under one master heading (with the shared
#    label) and new secondary headings are created based on the associated
#    CLIENT or EMPLOYER information.
#

package Resume;		# NB: Extends Resume, not Resume::Functionalize!

use strict;
use Resume;


#----------------------------------------------------------------------
# Add a (top-level) functional SECTION node to a section group
# (private instance method)

sub _fn_add_SECTION
{
	my ($self, $work, $node) = @_;
	my $label = $node->{'attribute'}{'label'};

	if (defined($label))
	{
		#
		# Save the most recent client/employer for use as the
		# sub-heading.
		#
		$node->{'employer'} = $work->{'emps'}[-1];

		#
		# Add this label to the section presentation order if
		# it hasn't been seen previously.
		#
		push(@{$work->{'section_order'}}, $label)
		  unless ($work->{'section'}{$label});

		#
		# Add this section to the group for this label.
		#
		push(@{$work->{'section'}{$label}}, $node);
	}
}

#----------------------------------------------------------------------
# Scan a CLIENT or EMPLOYER node (private instance method)

sub _fn_scan_EMPLOYER
{
	my ($self, $work, $node) = @_;
	my $sect_count = 0;
	my @retain;

	#
	# Save node on EMPLOYER or CLIENT stack
	#
	push(@{$work->{'emps'}}, $node)
	  if ($node->{'type'} eq 'b-tag' && $node->{'attribute'}{'name'});

	#
	# Use <B> instead of <Hn> in work history.
	#
	$node->{'work_fmt'} = '-h';

	foreach (@{$node->{'content'} || []})
	{
		#
		# Move any sections into section groups by label.
		#
		if (dispatch($_,
		'b-tag' => { 'section' =>
		  sub { $self->_fn_add_SECTION($work, $_); }, }))
		{
			++$sect_count;
			next;
		}

		#
		# Scan any nested CLIENT or EMPLOYER tags.
		#
		dispatch($_,
		'b-tag' =>
		{
			'client'  =>
			  sub { $self->_fn_scan_EMPLOYER($work, $_); },
			'employer'  =>
			  sub { $self->_fn_scan_EMPLOYER($work, $_); },
		},
		'tag' =>
		{
			'client' =>
			  sub { $self->_fn_scan_EMPLOYER($work, $_); },
			 'employer'  =>
			  sub { $self->_fn_scan_EMPLOYER($work, $_); },
		},
		);
		push(@retain, $_);
	}

	pop(@{$work->{'emps'}})
	  if ($node->{'type'} eq 'b-tag' && $node->{'attribute'}{'name'});

	if ($node->{'type'} eq 'b-tag')
	{
		#
		# If we deleted any SECTIONs (which would have forced
		# a work header display), make sure that the work header
		# will still get displayed.
		#
		unshift(@retain, {
			'type' => 'tag',
			'tag' => 'employer',
			'attribute' =>
			{
				'display' => 1,
			},
		  })
		  if ($sect_count);

		#
		# Retain any non-SECTION content in the work history.
		#
		$node->{'content'} = \@retain;
	}
}

#----------------------------------------------------------------------
# Functionalize a work SECTION node (private instance method)

sub _functionalize_work
{
	my ($self, $node) = @_;
	my $attr = $node->{'attribute'};
	my $label;

	$node->{'emps'} = [];		# CLIENT/EMPLOYER stack

	#
	# Create an employment history content node.
	#
	my $hist = $node->{'history'} =
	{
		'type' => 'b-tag',
		'tag' => 'section',
		'attribute' =>
		{
			'label' => $attr->{'history_label'} ||
			  $self->{'dict'}{'Work History'} ||
			  'Work History',
			'type' => 'history',
		},
		'content' => [],
	};

	#
	# We will be grouping top-level SECTIONs by label and the
	# order in which their labels are encountered.
	#
	$node->{'section'} = {};
	$node->{'section_order'} = [];

	#
	# Scan the sub-tree to find the interesting nodes.
	#
	foreach (@{$node->{'content'}})
	{
		#
		# <SECTION label='...'/> tags may be used to set the
		# section presentation order. Since they are pre-filtered,
		# they can be rearranged using target-specific order
		# ranking.
		#
		if ($_->{'type'} eq 'tag' && $_->{'tag'} eq 'section')
		{
			$label = $_->{'attribute'}{'label'} || '';

			if ($label ne '' && !$node->{'section'}{$label})
			{
				#
				# This is the first occurence of this label.
				# Add it to the presentation order.
				#
				push(@{$node->{'section_order'}}, $label);
				$node->{'section'}{$label} = [];
			}

			# There is no additional value to this node.
			next;
		}

		#
		# Visit CLIENT and EMPLOYER nodes to get their top-level
		# SECTIONs (and any nested CLIENT or EMPLOYER nodes).
		#
		$self->_fn_scan_EMPLOYER($node, $_)
		  if ($_->{'type'} eq 'b-tag' &&
		  ($_->{'tag'} eq 'employer' || $_->{'tag'} eq 'client'));

		#
		# Move this content into the work history.
		#
		push(@{$hist->{'content'}}, $_);
	}
	
	#
	# Determine if we're supposed to insert or append the
	# work history into the document (or neither).
	#
	my $hist_pos = $attr->{'history'} || 'after';
	if ($hist_pos eq 'before')
	{
		$self->_relink_content($hist);
		$self->_insert_node($node, $hist);
	}
	elsif ($hist_pos eq 'after')
	{
		$self->_relink_content($hist);
		$self->_append_node($node, $hist);
	}

	#
	# Reassemble the work SECTION content with the SECTION groups as
	# the primary content and CLIENT/EMPLOYER sub-headings below them.
	#
	$node->{'content'} = [];
	foreach $label (@{$node->{'section_order'}})
	{
		#
		# Clone a new node from the first node of each SECTION
		# group, using the entire SECTION group as its content.
		#
		my $new_node =
		{
			%{$node->{'section'}{$label}[0]},
			'content' => $node->{'section'}{$label},
		};


		#
		# Turn each SECTION node into a clone of the original
		# EMPLOYER/CLIENT node that enclosed it, but retaining
		# the SECTION's content and always using EMPLOYER.
		#
		foreach (@{$new_node->{'content'}})
		{
			%$_ = (
				%{$_->{'employer'}},
				'tag' => 'employer',
				'content' => $_->{'content'},
				'work_fmt' => 'p',
			);
		}

		#
		# Patch up the node links and add to the work SECTION
		# content.
		#
		$self->_relink_content($new_node);
		push(@{$node->{'content'}}, $new_node);
	}
}

#----------------------------------------------------------------------
# Convert a chronological resume into a functional resume
# (public instance method; main entry point)

sub functionalize
{
	my ($self) = @_;

	$self->_filter_content();

	my $work = $self->find_node(sub
	{
		my ($node) = @_;

		return ($node->{'type'} eq 'b-tag' &&
		  $node->{'tag'} eq 'section' &&
		  ($node->{'attribute'}{'type'} || '') eq 'work');
	});

	$self->_functionalize_work($work) if ($work);
}

1;
